from __future__ import annotations

import itertools
from enum import Enum
from functools import partial
from typing import Any, Protocol

from py_straight_skeleton.constants import NONCOLLINEARITY_EPSILON_DEGS, TIGHT_COLLINEARITY_EPSILON_DEGS
from py_straight_skeleton.vector_math import Line2, Orientation, Ray2, Vector2


class PolygonMathTracerProtocol(Protocol):
    """Protocol to receive notifications of key polygon math operations, specially edge cases."""

    def on_compute_from_edges_are_opposite(self, vertex_info: VertexInfo):
        """Called when the edges that spawn a new vertex are parallel and face opposite directions (anti-parallel)."""
        ...

    def on_compute_from_edges_are_aligned(self, vertex_info: VertexInfo):
        """Called when the edges that spawn a new vertex are parallel and face the same direction (aligned)."""
        ...

    def on_lav_infinite_loop_detected(self, lav: LAV):
        """Called when we iterate a LAV with an infinite loop."""
        ...

    def on_degenerate_lav_detected(self, lav: LAV):
        """Called when a LAV is detected to be degenerate (all vertices are the same)."""
        ...


class NoopPolygonMathTracer(PolygonMathTracerProtocol):
    """Implementation of the PolygonMathTracerProtocol that does nothing on each notification."""

    def on_compute_from_edges_are_opposite(self, vertex_info: VertexInfo):
        pass

    def on_compute_from_edges_are_aligned(self, vertex_info: VertexInfo):
        pass

    def on_lav_infinite_loop_detected(self, lav: LAV):
        pass

    def on_degenerate_lav_detected(self, lav):
        pass


# Global tracer that can be replaced by the user to receive key events for debugging and logging.
GLOBAL_POLYGON_MATH_TRACER: PolygonMathTracerProtocol = NoopPolygonMathTracer()


def set_global_polygon_math_tracer(tracer: PolygonMathTracerProtocol):
    """Set a custom global tracer for polygon math operations."""
    global GLOBAL_POLYGON_MATH_TRACER
    GLOBAL_POLYGON_MATH_TRACER = tracer


class InsidePolygonCheckPolicy(Enum):
    """Policy to follow when we check if a vector is inside a LAV, specifically around parallel segments."""

    RELAXED = 0  # Checks will generally accept cases if any condition passes.
    STRICT = 1  # Checks will generally accept cases if all conditions pass.
    RAISE_ERROR = 2  # Checks will raise an error, failing execution, for parallel cases (mostly for debugging).


def is_vector_in_segment_polygon(
    prev_segment: Vector2,
    next_segment: Vector2,
    candidate_vector: Vector2,
    policy: InsidePolygonCheckPolicy = InsidePolygonCheckPolicy.RAISE_ERROR,
    collinearity_epsilon_degrees: float = TIGHT_COLLINEARITY_EPSILON_DEGS,
) -> bool:
    """Check if a candidate vector is inside the polygon that the two segments belong to. The polygon, by convention,
    is to the left of the segments. Note that here "polygon" may refer to the polygon defined by a LAV, not necessarily
    the initial polygon.

    Args:
        prev_segment: Segment that arrives to the vertex in the polygon we are checking.
        next_segment: Segment that leaves the vertex in the polygon we are checking.
        candidate_vector: Vector whose direction we want to check, and that is considered as starting on the same
            vertex as the segments.
        policy: Policy to follow when comparing against parallel segments. See InsidePolygonCheckPolicy.

    Raises:
        RuntimeError: If the policy is set to RAISE_ERROR and the segments are (parallel) aligned or opposite.
        ValueError: If the specified policy is not recognized.

    Returns:
        True if the candidate vector is inside the left polygon defined by the two given segments, False otherwise.
    """
    bis_vs_p_segment = Orientation.of_vec_with_respect_to(
        vec=candidate_vector, wrt=prev_segment, collinearity_epsilon_degrees=collinearity_epsilon_degrees
    )
    bis_vs_n_segment = Orientation.of_vec_with_respect_to(
        vec=candidate_vector, wrt=next_segment, collinearity_epsilon_degrees=collinearity_epsilon_degrees
    )
    segment_ori = Orientation.of_vec_with_respect_to(
        vec=next_segment, wrt=prev_segment, collinearity_epsilon_degrees=collinearity_epsilon_degrees
    )
    # TODO Need to review whether the inclusive checks here should allow opposites. Add unit-tests.
    if segment_ori.is_left_exclusive():
        # Convex angle
        # The polygon is to the left if bisector is Left-Left with respect to our segments.
        is_polygon_left = bis_vs_p_segment.is_left_inclusive() and bis_vs_n_segment.is_left_inclusive()
    elif segment_ori.is_right_exclusive():
        # Reflex angle
        # The polygon is to the left if the bisector is not Right-Right with respect to our segments.
        is_polygon_left = bis_vs_p_segment.is_left_inclusive() or bis_vs_n_segment.is_left_inclusive()
    elif segment_ori.is_aligned():
        # This is like the convex angle, however it is possible that the bisector is between our segments, failing
        # the Left-Left check. Pick which vector to check based on the policy.
        if policy == InsidePolygonCheckPolicy.RELAXED:
            # Relaxed, it is inside as long as it is with respect to the most open segment.
            is_polygon_left = bis_vs_p_segment.is_left_inclusive() or bis_vs_n_segment.is_left_inclusive()
        elif policy == InsidePolygonCheckPolicy.STRICT:
            # Strict, it is inside only if it is inside both segments.
            is_polygon_left = bis_vs_p_segment.is_left_inclusive() and bis_vs_n_segment.is_left_inclusive()
        elif policy == InsidePolygonCheckPolicy.RAISE_ERROR:
            raise RuntimeError("Parallel segments, can't determine if the bisector is inside the left-polygon.")
        else:
            raise ValueError(f"Unknown policy {policy}")
    elif segment_ori.is_opposite():
        if policy == InsidePolygonCheckPolicy.RELAXED:
            # Relaxed, it is inside as long as it is antiparallel to the previous edge (the one that goes into
            # the vertex), or parallel to the next edge (the one that goes out of the vertex).
            is_polygon_left = bis_vs_p_segment.is_opposite() or bis_vs_n_segment.is_aligned()
        elif policy == InsidePolygonCheckPolicy.STRICT:
            # Strict, it is inside as long as it is antiparallel to the previous edge (the one that goes into
            # the vertex), and parallel to the next edge (the one that goes out of the vertex).
            is_polygon_left = bis_vs_p_segment.is_opposite() and bis_vs_n_segment.is_aligned()
        elif policy == InsidePolygonCheckPolicy.RAISE_ERROR:
            raise RuntimeError("Antiparallel segments, can't determine if the bisector is inside the left-polygon.")
        else:
            raise ValueError(f"Unknown policy {policy}")
    return is_polygon_left


def is_point_inside_polygon_edges(edges: list[Edge], point2d: Vector2) -> bool:
    """Determine if a point is inside a polygon defined by its edges using the ray-casting algorithm.

    Note holes are purposely ignored, as they are not relevant for this check. This method returns true even if
    the point is inside a hole.

    Args:
        edges: List of edges that form the outside of the polygon. Holes are not relevant for this check.
        point2d: Point to check if it is inside the polygon.

    Returns:
        True if the point is inside the polygon exterior, defined by its edges, False otherwise.
    """
    x, y = point2d
    inside = False

    for edge in edges:
        (x1, y1), (x2, y2) = edge.start_vertex.position.coords, edge.end_vertex.position.coords

        # Check if the point is within the vertical range of the edge
        if min(y1, y2) < y <= max(y1, y2):
            # Calculate the intersection of the edge with the ray extending from the point to the right
            if x1 != x2:  # Avoid division by zero for vertical lines
                x_intersect = x1 + (y - y1) * (x2 - x1) / (y2 - y1)
            else:
                x_intersect = x1  # The edge is vertical

            # If the intersection is to the right of the point, toggle the 'inside' state
            if x <= x_intersect:
                inside = not inside

    return inside


class DegenerateLAVError(Exception):
    """Custom exception for degenerate LAVs (all points in the LAV are the same)."""

    pass


class VertexNotComputedError(Exception):
    """Custom exception when we try to access properties of a vertex that were not yet computed."""

    pass


class VertexInfo:
    """Extended information about a Vertex in a LAV."""

    __id_counter = itertools.count()

    @classmethod
    def reset_count(cls):
        cls.__id_counter = itertools.count()

    def __init__(self, position: Vector2, lav_ptr: LAV, split_dummy_vertex: VertexInfo | None = None):
        """Initializes a new VertexInfo at the given position, with a pointer to its containing LAV.

        Many of the attributes accessed through properties require computation, and are not available until the
        computation is executed. See `compute_from_edges`.

        Args:
            position: Position of the vertex.
            lav_ptr: Pointer to the LAV that contains this vertex.
            split_dummy_vertex: Optional dummy vertex that vertices created by split events may have. This is used to
                link arcs to the same vertex (the dummy vertex), even when the arcs reference one of the split vertices.
        """
        self._auto_id = next(self.__id_counter)

        self.position = position
        self.lav_ptr = lav_ptr
        self.processed = False
        self.split_dummy_vertex = split_dummy_vertex

        # Pointers to the original polygon edges.
        self.prev_orig_edge: Edge | None = None
        self.next_orig_edge: Edge | None = None
        # Linked list pointers.
        self.__next_vertex: VertexInfo | None = None
        self.__prev_vertex: VertexInfo | None = None

        # Private attributes (some need to be computed later).
        self.__computed: bool = False
        self.__bisector_direction: Vector2 | None = None
        self.__is_reflex: bool | None = None

        self._pending_split_events: list[Any] | None = None
        self._pending_split_events_computed = False

    def __hash__(self):
        """For comparison. Note that only the id is taken into account, so two vertex infos at the same location are
        considered different."""
        return hash(self._auto_id)

    def __repr__(self):
        return (
            f"VertexInfo(id={self._auto_id}"
            + f", position={self.position}"
            + f", lav={self.lav_ptr._auto_id}"
            + f", processed={self.processed}"
            + f", reflex={self.is_reflex if self.__computed else 'N/A'}"
            + f", [prev={self.prev_vertex._auto_id if self.__prev_vertex is not None else 'None'}"
            + f", next={self.next_vertex._auto_id if self.__next_vertex is not None else 'None'}]"
        )

    @property
    def next_vertex(self) -> VertexInfo:
        """Returns the next vertex in the LAV."""
        if self.__next_vertex is None:
            raise RuntimeError("next_vertex not set.")
        return self.__next_vertex

    @next_vertex.setter
    def next_vertex(self, value: VertexInfo):
        self.__next_vertex = value

    @property
    def prev_vertex(self) -> VertexInfo:
        """Returns the previous vertex in the LAV."""
        if self.__prev_vertex is None:
            raise RuntimeError("prev_vertex not set.")
        return self.__prev_vertex

    @prev_vertex.setter
    def prev_vertex(self, value: VertexInfo):
        self.__prev_vertex = value

    @property
    def pending_split_events(self) -> list[Any]:
        """Returns the pending split events for this vertex."""
        if not self._pending_split_events_computed:
            raise RuntimeError("Pending split events not computed.")
        assert self._pending_split_events is not None  # for mypy, since we set it when we compute it.
        return self._pending_split_events

    @pending_split_events.setter
    def pending_split_events(self, split_events: list[Any]):
        """Sets the pending split events for this vertex, flagging as computed."""
        if self._pending_split_events_computed:
            raise RuntimeError("Pending split events were already computed.")
        self._pending_split_events = split_events
        self._pending_split_events_computed = True

    @property
    def pending_split_events_computed(self) -> bool:
        """Returns whether the pending split events have been computed."""
        return self._pending_split_events_computed

    @property
    def is_reflex(self) -> bool:
        """Returns whether the vertex is reflex or not. The vertex info must be computed before accessing this property."""
        if not self.__computed:
            raise VertexNotComputedError()
        assert self.__is_reflex is not None  # for mypy, since we set it when we compute it.
        return self.__is_reflex

    # TODO I think `force_reflex` can be removed. See note in the caller.
    def force_reflex(self, value: bool):
        """This allows setting the reflex property after it has been auto-computed.

        Although not ideal, turns out to be necessary for edge events, which can accidentally detect reflex vertices
        when they are not. We should fix that underlying issue, but for now this is a workaround."""
        if not self.__computed:
            raise VertexNotComputedError()
        self.__is_reflex = value

    @property
    def bisector_direction(self) -> Vector2:
        """Returns the direction of the bisector of the vertex. The vertex info must be computed before accessing this property."""
        if not self.__computed:
            raise VertexNotComputedError()
        return self.__bisector_direction

    def bisector_ray(self):
        """Returns a Ray2 starting at the vertex and pointing in the same direction as the vertex's bisector."""
        return Ray2(self.position, self.bisector_direction)

    def get_prev_vertex_safe(self) -> VertexInfo:
        """Iterate our linked vertices backwards until we find one that is not overlapping with us.

        Raises:
            RuntimeError: If an infinite loop is detected.
            DegenerateLAVError: If all vertices are overlapping, and we can't find a safe previous vertex.

        Returns:
            The first vertex that is not overlapping with the vertex, found iterating backwards the linked list / LAV.
        """
        # Iterate vertices until we find one that is not overlapping with us (backward direction).
        visited = set()
        prev_ptr = self.prev_vertex
        visited.add(prev_ptr)
        while prev_ptr.position == self.position:
            prev_ptr = prev_ptr.prev_vertex
            # We might want to notify tracers if these happen.
            if prev_ptr in visited:
                raise RuntimeError("Infinite loop detected in `get_prev_vertex_safe`.")
            if prev_ptr == self:
                raise DegenerateLAVError("All vertices are overlapping, can't get a safe prev vertex.")
            visited.add(prev_ptr)
        return prev_ptr

    def get_next_vertex_safe(self) -> VertexInfo:
        """Iterate our linked vertices forward until we find one that is not overlapping with us.

        Raises:
            RuntimeError: If an infinite loop is detected.
            DegenerateLAVError: If all vertices are overlapping, and we can't find a safe next vertex.

        Returns:
            The first vertex that is not overlapping with the vertex, found iterating forward the linked list / LAV.
        """
        # Iterate vertices until we find one that is not overlapping with us (forward direction).
        visited = set()
        next_ptr = self.next_vertex
        visited.add(next_ptr)
        while next_ptr.position == self.position:
            next_ptr = next_ptr.next_vertex
            # We might want to notify tracers if these happen.
            if next_ptr in visited:
                raise RuntimeError("Infinite loop detected in `get_next_vertex_safe`.")
            if next_ptr == self:
                raise DegenerateLAVError("All vertices are overlapping, can't get a safe next vertex.")
            visited.add(next_ptr)
        return next_ptr

    def _is_self_bisector_inside_lav(self) -> bool:
        """Return whether our computed bisector is pointing towards the inside the LAV, comparing our previous
        and next safe sements (that is, the first ones whose other point does not overlap with this vertex).

        Returns:
            True if the bisector is pointing towards inside the LAV, False otherwise.
        """
        try:
            prev_segment = (self.position - self.get_prev_vertex_safe().position).normalized()
            next_segment = (self.get_next_vertex_safe().position - self.position).normalized()
        except DegenerateLAVError:
            return False  # LAV is a point, no bisector will be inside.

        return is_vector_in_segment_polygon(prev_segment, next_segment, candidate_vector=self.bisector_direction)

    def compute_from_edges(self) -> None:
        """Compute the bisector of the vertex and whether it is reflex, based on the edges that spawn it.

        Raises:
            RuntimeError: If the vertex's properties were already computed.
        """
        if self.__computed:
            raise RuntimeError("Vertex was already computed.")

        if self.prev_orig_edge is None or self.next_orig_edge is None:
            raise RuntimeError("Edges have not been set in the vertex.")

        in_edge = self.prev_orig_edge.direction
        out_edge = self.next_orig_edge.direction
        in_edge_inv = -in_edge

        # Figuring out the direction of the bisector from the creating edges is not as trivial as it seems. In
        # particular, there are edge cases around collinear edges, which can happen in different ways for original
        # vertices and non-original skeleton nodes.
        is_in_original_prev_edge = self.prev_orig_edge.has_endpoint(self)
        is_in_original_next_edge = self.next_orig_edge.has_endpoint(self)
        assert is_in_original_prev_edge == is_in_original_next_edge, "Vertices should belong to both or neither edge."
        is_original_vertex = is_in_original_prev_edge

        ori = Orientation.of_vec_with_respect_to(
            vec=out_edge, wrt=in_edge, collinearity_epsilon_degrees=TIGHT_COLLINEARITY_EPSILON_DEGS
        )
        if ori.is_opposite():
            # When they are opposite, we are unsure if the LAV is on the convex side or the reflex side. Attempt
            # to check very strictly against the LAV. If only one of the bisectors is inside, then that is the one we
            # want. If both are inside, we are still confused.
            bisector_pos = (in_edge_inv + out_edge).normalized()
            bisector_neg = -bisector_pos
            try:
                # Note that in/out segments are equal to the edges for the original vertices.
                in_segment = (self.position - self.get_prev_vertex_safe().position).normalized()
                out_segment = (self.get_next_vertex_safe().position - self.position).normalized()
            except DegenerateLAVError as e:
                # If the LAV is collapsed into a point, no bisector will be inside. This should not happen.
                GLOBAL_POLYGON_MATH_TRACER.on_degenerate_lav_detected(lav=self.lav_ptr)
                raise e
            is_in_pos = is_vector_in_segment_polygon(
                prev_segment=in_segment,
                next_segment=out_segment,
                candidate_vector=bisector_pos,
                collinearity_epsilon_degrees=NONCOLLINEARITY_EPSILON_DEGS,
            )
            is_in_neg = is_vector_in_segment_polygon(
                prev_segment=in_segment,
                next_segment=out_segment,
                candidate_vector=bisector_neg,
                collinearity_epsilon_degrees=NONCOLLINEARITY_EPSILON_DEGS,
            )
            if is_in_neg == is_in_pos:
                # See if the less relaxed version can tell them apart.
                is_in_pos = is_vector_in_segment_polygon(
                    prev_segment=in_segment,
                    next_segment=out_segment,
                    candidate_vector=bisector_pos,
                    collinearity_epsilon_degrees=TIGHT_COLLINEARITY_EPSILON_DEGS,
                )
                is_in_neg = is_vector_in_segment_polygon(
                    prev_segment=in_segment,
                    next_segment=out_segment,
                    candidate_vector=bisector_neg,
                    collinearity_epsilon_degrees=TIGHT_COLLINEARITY_EPSILON_DEGS,
                )
            if is_in_neg == is_in_pos:
                raise RuntimeError("Can't determine bisector direction for opposite edges.")

            if is_in_pos:
                bisector = bisector_pos
                is_reflex = False
            else:
                bisector = bisector_neg
                is_reflex = True
            GLOBAL_POLYGON_MATH_TRACER.on_compute_from_edges_are_opposite(vertex_info=self)
        elif ori.is_aligned():
            # If the two edges are pointing in the same direction, their desired bisector is perpendicular to the edges,
            # and towards their left (inwards wrt the polygon).
            bisector = Vector2(-in_edge.y, in_edge.x).normalized()
            is_reflex = False
            GLOBAL_POLYGON_MATH_TRACER.on_compute_from_edges_are_aligned(vertex_info=self)
        else:
            # If the point is part of the edges, we are in the vanilla case.
            if is_original_vertex:
                bisector = (in_edge_inv + out_edge).normalized()
                is_reflex = ori.is_right_exclusive()
                if is_reflex:
                    bisector = -bisector
            else:
                bisector = (in_edge_inv + out_edge).normalized()
                is_reflex = ori.is_right_exclusive()

        # Set final values.
        self.__is_reflex = is_reflex
        self.__bisector_direction = bisector.normalized()
        self.__computed = True


class Edge:
    """Information about an edge in the original polygon."""

    __id_counter = itertools.count()

    @classmethod
    def reset_count(cls):
        cls.__id_counter = itertools.count()

    def __init__(self, start_vertex: VertexInfo, end_vertex: VertexInfo, is_original_edge: bool = True):
        """Initializes a new Edge between two vertices. The edge is automatically set on the vertices, if this
        edge is an original edge. Direction of the edge is from start to end.

        Args:
            start_vertex: Start vertex of the edge.
            end_vertex: End vertex of the edge.
            is_original_edge: Whether this is an edge in the original polygon.
        """
        self._auto_id = next(self.__id_counter)

        self.start_vertex = start_vertex
        self.end_vertex = end_vertex
        self.direction = (end_vertex.position - start_vertex.position).normalized()

        # Automatically set this edge on the vertices.
        if is_original_edge:
            start_vertex.next_orig_edge = self
            end_vertex.prev_orig_edge = self

    def __repr__(self):
        return f"Edge(id={self._auto_id}, {self.start_vertex.position}, {self.end_vertex.position})"

    def has_endpoint(self, vertex: VertexInfo) -> bool:
        """Return whether the edge has the given vertex as one of its endpoints."""
        return vertex == self.start_vertex or vertex == self.end_vertex

    @property
    def start_p(self):
        """Return the start vertex's position."""
        return self.start_vertex.position

    @property
    def end_p(self):
        """Return the end vertex's position."""
        return self.end_vertex.position

    @property
    def center_p(self):
        """Return the center point of the edge."""
        return (self.start_p + self.end_p) / 2

    def distance_point_to_edge_line(self, point: Vector2, signed: bool = False) -> float:
        """Return the distance from the point to the infinite line holding this edge, optionally signed/unsigned.

        Args:
            point: Point to compute the distance to the edge.
            signed: Whether to return the signed distance or the absolute distance.

        Returns:
            The distance from the point to the infinite line holding this edge.
        """
        # Convert the edge to a line and then compute the distance.
        line = Line2(self.start_p, self.direction)
        d = line.signed_distance(point)
        return d if signed else abs(d)

    def is_point_in_edge_region(
        self,
        point: Vector2,
        *,
        allow_inclusive: bool = True,
        collinearity_epsilon_degrees: float = TIGHT_COLLINEARITY_EPSILON_DEGS,
    ) -> bool:
        """Return whether the given point is inside the region defined by this edge and the bisectors of its
        vertices. The region is to the left of the edge, and the bisectors are considered as rays starting at the
        vertices.

        Note on `allow_inclusive`: Conceptually the checks should be inclusive, however some skeletons may fail
        depending on the collinearity epsilon values if inclusivity is enabled (specifically with eps > 1 deg).

        Args:
            point: Point to check if it is inside the region defined by this edge and the bisectors of its vertices.
            allow_inclusive: Whether to consider points on the boundaries as inside the region.

        Returns:
            True if the point is inside the region, False otherwise
        """
        # Classify the point based on its location relative to the boundaries.
        start_p, start_bis = self.start_vertex.position, self.start_vertex.bisector_direction
        end_p, end_bis = self.end_vertex.position, self.end_vertex.bisector_direction

        left_check = (
            partial(Orientation.is_left_inclusive, allow_aligned=True, allow_opposite=False)
            if allow_inclusive
            else Orientation.is_left_exclusive
        )
        right_check = (
            partial(Orientation.is_right_inclusive, allow_aligned=True, allow_opposite=False)
            if allow_inclusive
            else Orientation.is_right_exclusive
        )

        # Special case for the start and end points, since otherwise we can't normalize their directions.
        if point == start_p or point == end_p:
            return allow_inclusive

        # Needs to be:
        # a) right or on top of the start bisector.
        start_ori = Orientation.of_vec_with_respect_to(
            vec=(point - start_p), wrt=start_bis, collinearity_epsilon_degrees=collinearity_epsilon_degrees
        )
        on_start_bis_right = right_check(start_ori)
        # b) left or on top of the edge.
        edge_ori = Orientation.of_vec_with_respect_to(
            vec=(point - start_p), wrt=self.direction, collinearity_epsilon_degrees=collinearity_epsilon_degrees
        )
        on_edge_left = left_check(edge_ori)
        # c) left or on top of the end bisector.
        end_ori = Orientation.of_vec_with_respect_to(
            vec=(point - end_p), wrt=end_bis, collinearity_epsilon_degrees=collinearity_epsilon_degrees
        )
        on_end_bis_left = left_check(end_ori)

        # The point is inside the region if it's on the correct side of all three boundaries.
        return on_start_bis_right and on_edge_left and on_end_bis_left


class LAV:
    """List of active vertices."""

    # Automatic counter to assign unique ids to each LAV.
    __id_counter = itertools.count()

    @classmethod
    def reset_count(cls):
        cls.__id_counter = itertools.count()

    def __init__(self, head: VertexInfo | None = None):
        """Initialize the LAV with an optional head vertex.

        Args:
            head: Initial vertex of the LAV (if any).
        """
        self._auto_id = next(self.__id_counter)

        self.head = head

    def __hash__(self) -> int:
        """For storage in the set of LAVs. Note that only the id is taken into account."""
        return hash(self._auto_id)

    def __eq__(self, other: object) -> bool:
        """Compares two LAVS, based exclusively on their unique ids, regardless of other values."""
        if not isinstance(other, LAV):
            return NotImplemented
        return self._auto_id == other._auto_id

    def __iter__(self):
        """Iterate over the vertices in the LAV, checking for (infinite) loops."""
        if self.head is None:
            return
        visited = set()  # Should not be needed, but the code is setup for tracing unexpected cases.
        current = self.head

        # We do unbounded iteration since we check for loops. Otherwise we might want to set up a limit.
        while True:
            yield current
            visited.add(current)
            current = current.next_vertex
            if current == self.head:
                break
            if current in visited:
                GLOBAL_POLYGON_MATH_TRACER.on_lav_infinite_loop_detected(self)
                raise RuntimeError(f"Infinite loop in LAV. Visited {len(visited)} vertices.")

    def append_vertex_at(self, point: Vector2) -> None:
        """Creates a new vertex info and adds it to the tail of the LAV, at the specified point.

        Args:
            point: 2D point to add a vertex at. The other vertex properties are not yet computed, since they require
                additional information.
        """
        new_vertex = VertexInfo(position=point, lav_ptr=self)
        if self.head is None:
            self.head = new_vertex
            self.head.next_vertex = self.head
            self.head.prev_vertex = self.head
        else:
            last_vertex = self.head.prev_vertex
            # Insert the new vertex between the head and the last
            self.head.prev_vertex = new_vertex
            last_vertex.next_vertex = new_vertex
            new_vertex.prev_vertex = last_vertex
            new_vertex.next_vertex = self.head



"""Module with constant definitions to share across code."""

# Note: It may be slightly dangerous to not specify units, since epsilons may need scaling depending on polygon units.
# DISTANCE_EPSILON: Epsilon to compare positions, distances, etc.
DISTANCE_EPSILON = 1e-5

# TIME_EPSILON: Used to compare event times that may have floating-point imprecision (for example to identify
# simultaneous events). Note that in skeleton computation time is equivalent to distance, and, in general, given
# the current implementation, it should probably be <= DISTANCE_EPSILON.
TIME_EPSILON = DISTANCE_EPSILON

# I have empirically seen `v.dot(v)` to fail collinearity checks if the epsilon is <=0.2
# >>> math.degrees(math.acos(1 - 1e-5))
# 0.25623472915884415
# TIGHT_COLLINEARITY_EPSILON_DEGS: Used when we want to check that vectors are almost collinear.
TIGHT_COLLINEARITY_EPSILON_DEGS = 0.26
# NONCOLLINEARITY_EPSILON_DEGS: Used when we are going to trust the precision, even if seemingly collinear.
NONCOLLINEARITY_EPSILON_DEGS = 0.0

# DET_EPSILON: Epsilon to compare determinants during intersection computations from 2d cross products, in order
# to prevent division by 0.
DET_EPSILON = 1e-5



# read version from installed package
from importlib.metadata import version

__version__ = version("py_straight_skeleton")

from .algorithm import PointList, StraightSkeletonAlgorithm
from .skeleton import Skeleton


def compute_skeleton(exterior: PointList, holes: list[PointList]) -> Skeleton:
    """Computes the straight skeleton of the polygon defined by the given exterior outline and interior holes.

    Args:
        exterior: List of vertices that define the exterior of the polygon. Must be provided in COUNTER-CLOCKWISE,
            with +X-right and +Y-up.
        holes: List of holes, each as a list of vertices that define the hole. Must be provided in CLOCKWISE order,
            with +X-right and +Y-up (opposite of the exterior vertices).

    Returns:
        The straight skeleton of the polygon defined by the exterior and holes.
    """
    algorithm = StraightSkeletonAlgorithm()
    return algorithm.compute_skeleton(exterior=exterior, holes=holes)



from __future__ import annotations

import math
from enum import Enum

from py_straight_skeleton.constants import DET_EPSILON, DISTANCE_EPSILON, TIGHT_COLLINEARITY_EPSILON_DEGS


class Vector2:
    """2D vector/point class."""

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __getitem__(self, index):
        return (self.x, self.y)[index]

    def __repr__(self):
        return f"Vector2({self.x:.03f}, {self.y:.03f})"

    @property
    def coords(self):
        return (self.x, self.y)

    def __add__(self, other: Vector2) -> Vector2:
        return Vector2(self.x + other.x, self.y + other.y)

    def __sub__(self, other: Vector2) -> Vector2:
        return Vector2(self.x - other.x, self.y - other.y)

    def __mul__(self, scalar: float) -> Vector2:
        return Vector2(self.x * scalar, self.y * scalar)

    def __truediv__(self, scalar: float | int) -> Vector2:
        return Vector2(self.x / scalar, self.y / scalar)

    def __floordiv__(self, scalar: float | int) -> Vector2:
        return Vector2(self.x // scalar, self.y // scalar)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Vector2):
            return NotImplemented
        return abs(self.x - other.x) < DISTANCE_EPSILON and abs(self.y - other.y) < DISTANCE_EPSILON

    def __neg__(self) -> Vector2:
        return Vector2(-self.x, -self.y)

    def dot(self, other: Vector2) -> float:
        return self.x * other.x + self.y * other.y

    def cross(self, other: Vector2) -> float:
        return self.x * other.y - self.y * other.x

    def length(self) -> float:
        return (self.x**2 + self.y**2) ** 0.5

    def normalized(self) -> Vector2:
        """Returns a copy of this vector, normalized by its length. It does not normalize itself."""
        return self / self.length()


class Orientation(Enum):
    """Orientation of a vector with respect to another vector. Note that here we treat vectors as direction vectors. The
    same orientation may not hold if the vectors are translated to their respective origins, for example when comparing
    rays.

        Note: It is encouraged to use methods for comparisons instead of directly using == or != operators.
    """

    _LEFT = 0
    _RIGHT = 1
    _ALIGNED = 2
    _OPPOSITE = 3

    def is_left_exclusive(self) -> bool:
        """The orientation signifies left, but does not include collinear cases."""
        return self == self._LEFT

    def is_left_inclusive(self, allow_aligned: bool = True, allow_opposite: bool = True) -> bool:
        """The orientation signifies left, and includes collinear cases if allowed."""
        if self == self._LEFT:
            return True
        if self == self._ALIGNED:
            return allow_aligned
        if self == self._OPPOSITE:
            return allow_opposite
        return False

    def is_right_exclusive(self) -> bool:
        """The orientation signifies right, but does not include collinear cases."""
        return self == self._RIGHT

    def is_right_inclusive(self, allow_aligned: bool = True, allow_opposite: bool = True) -> bool:
        """The orientation signifies right, and includes collinear cases if allowed."""
        if self == self._RIGHT:
            return True
        if self == self._ALIGNED:
            return allow_aligned
        if self == self._OPPOSITE:
            return allow_opposite
        return False

    def is_aligned(self) -> bool:
        """The vectors are aligned, i.e. they are parallel and point in the same direction."""
        return self == self._ALIGNED

    def is_opposite(self) -> bool:
        """The vectors are opposite, i.e. they are (anti)-parallel pointing in opposite directions."""
        return self == self._OPPOSITE

    def is_collinear(self) -> bool:
        """The vectors are collinear, i.e. they are parallel/aligned or anti-parallel/opposite."""
        return self == self._ALIGNED or self == self._OPPOSITE

    @classmethod
    def of_vec_with_respect_to(
        cls,
        vec: Vector2,
        wrt: Vector2,
        *,
        collinearity_epsilon_degrees: float = TIGHT_COLLINEARITY_EPSILON_DEGS,
    ):
        """Computes and returns the orientation of `vec` with respect to `wrt`. Both vectors are normalized before
        computing the orientation, to guarantee that the result is consistent.

        Args:
            vec: Vector to compute orientation of.
            wrt: Vector to compute orientation with respect to.
            collinearity_epsilon_degrees: Epsilon value in degrees to consider vectors collinear. If the angle between
                the vectors is less than this value, they are considered collinear.

        Returns:
            _description_
        """
        vec = vec.normalized()
        wrt = wrt.normalized()

        dot = wrt.dot(vec)
        cross = wrt.cross(vec)
        angle_radians = math.atan2(cross, dot)
        angle_degrees = math.degrees(angle_radians)
        if abs(angle_degrees) < collinearity_epsilon_degrees:
            return cls._ALIGNED
        if abs(180 - abs(angle_degrees)) < collinearity_epsilon_degrees:
            return cls._OPPOSITE
        if angle_degrees > 0:
            return cls._LEFT
        return cls._RIGHT


class Ray2:
    """2D ray class, defined by an origin and a direction vector."""

    def __init__(self, origin: Vector2, direction: Vector2):
        """Initialize the ray, automatically normalizing the direction vector."""
        self.origin = origin
        self.direction = direction.normalized()

    def intersection_with_ray(
        self,
        other: Ray2,
        *,
        allow_forward_collinear: bool = True,
        allow_backward_collinear: bool = True,
        dist_epsilon: float = DISTANCE_EPSILON,
        collinearity_epsilon_deg: float = TIGHT_COLLINEARITY_EPSILON_DEGS,
    ) -> Vector2 | None:
        """Compute intersection point of this ray with another ray.

        Args:
            other: Ray to compute intersection with.
            allow_forward_collinear: Whether to allow finding an intersection point when the rays are facing the same
                direction, and the other ray's origin is ahead of this ray's origin. This is useful for bisectors to
                find intersections with other bisectors that share direction but whose vertex is ahead of ours.
            allow_backward_collinear: Whether to allow finding an intersection point when the rays are facing the same
                direction, and this ray's origin is ahead of the other ray's origin. This is useful for bisectors to
                find that their vertex is ahead and intersected by the other bisector.
            dist_epsilon: Epsilon to compare distances.
            collinearity_epsilon_deg: Epsilon in degrees to consider vectors collinear. If the angle between the vectors
                is less than this value, they are considered collinear.

        Raises:
            RuntimeError: If computations take unexpected values (this would signal an implementation error).

        Returns:
            The intersection point if it exists, according to the args, or None if the rays do not intersect.
        """
        # General case, compute the intersection point.
        # Calculate the determinant.
        det = self.direction.cross(other.direction)
        if abs(det) > DET_EPSILON:
            # Calculate the difference between origins.
            diff = other.origin - self.origin

            # Calculate parameters t and u.
            t = diff.cross(other.direction) / det
            u = diff.cross(self.direction) / det

            # Check if intersection point is on both rays.
            if t >= -dist_epsilon and u >= -dist_epsilon:
                # Calculate and return the intersection point
                return self.origin + self.direction * t
            else:
                # The intx
                return None

        # The code below handles the case when the rays are fairly collinear, but not precisely. In that case,
        # we detect approximate collinear intersections, if allowed by the caller.

        # If not allowed, we did not find a collision so far.
        if not allow_backward_collinear and not allow_forward_collinear:
            return None
        # If the origins are the same, then we also don't want to collide with ourselves.
        if self.origin == other.origin:
            return None

        # If allowed, check still the orientation (this could be optimized since we already have the det).
        ori = Orientation.of_vec_with_respect_to(
            vec=other.direction, wrt=self.direction, collinearity_epsilon_degrees=collinearity_epsilon_deg
        )
        if ori.is_aligned():
            # Compute whether the other ray is ahead/forward or behind/backward with respect to us.
            cos_epsilon = math.cos(math.radians(collinearity_epsilon_deg))
            self_to_other_dir = (other.origin - self.origin).normalized()
            self_to_other_cos = self_to_other_dir.dot(self.direction)
            has_foward_collinear = self_to_other_cos >= cos_epsilon
            has_backward_collinear = self_to_other_cos <= -cos_epsilon
            if allow_forward_collinear and has_foward_collinear:
                return other.origin
            if allow_backward_collinear and has_backward_collinear:
                return self.origin
            # No collinear intersection allowed or found.
            return None
        elif ori.is_opposite():
            # Anti-parallel rays do not intersect (arguable claim, but for our purposes they don't).
            return None

        # No intersection found, not collinear trick.
        return None


class Line2:
    """2D line class, defined by a point and a direction vector. Unlike a Ray, Lines are infinite and undirected."""

    def __init__(self, point: Vector2, direction: Vector2):
        self.point = point
        self.direction = direction.normalized()

    def intersection_with_ray(
        self, ray: Ray2, *, collinearity_epsilon_deg: float = TIGHT_COLLINEARITY_EPSILON_DEGS
    ) -> Vector2 | None:
        """Compute intersection point of this line with a ray.

        Args:
            ray: Ray to compute intersection with.
            collinearity_epsilon_deg: TBD

        Returns:
            The intersection point if it exists, or None if the line and ray do not intersect.
        """
        collinearity_sin_epsilon = math.sin(math.radians(collinearity_epsilon_deg))

        # Get determinant and check for parallelism.
        det = self.direction.cross(ray.direction)
        if abs(det) <= collinearity_sin_epsilon:
            return None

        # Compute the intersection point.
        diff = ray.origin - self.point
        u = diff.cross(self.direction) / det

        # Check if intersection point is on the ray.
        if u >= 0:
            ret = ray.origin + ray.direction * u
            assert ret == self.point + self.direction * (diff.cross(ray.direction) / det)
            return ret

        # No intersection if u < 0 (ray points away).
        return None

    def intersection_with_line(
        self, line: Line2, *, collinearity_epsilon_deg: float = TIGHT_COLLINEARITY_EPSILON_DEGS
    ) -> Vector2 | None:
        """Compute intersection point of this line with another line.

        Args:
            line: Line to compute intersection with.
            collinearity_epsilon_deg: TBD

        Returns:
            The intersection point if it exists, or None if the lines do not intersect.
        """
        collinearity_sin_epsilon = math.sin(math.radians(collinearity_epsilon_deg))

        # Get determinant and check for parallelism.
        det = self.direction.cross(line.direction)
        if abs(det) <= collinearity_sin_epsilon:
            return None

        # Compute the intersection point.
        diff = line.point - self.point
        t = diff.cross(line.direction) / det
        intersection_point = self.point + self.direction * t
        return intersection_point

    def signed_distance(self, point: Vector2) -> float:
        """Compute the signed distance of a point to this line. The sign indicates on which side of the line the point
        lies.

        Args:
            point: Point to compute distance to.

        Returns:
            The signed distance of the point to the line.
        """
        perp_vector = Vector2(-self.direction.y, self.direction.x)
        return perp_vector.dot(point - self.point)

    def find_point_with_min_distance_along_vector(
        self,
        point: Vector2,
        vector: Vector2,
        min_distance: float,
        *,
        collinearity_epsilon_deg: float = TIGHT_COLLINEARITY_EPSILON_DEGS,
    ) -> tuple[float, Vector2]:
        """Find a point along the given vector that is at a minimum distance from this line. In order to travel the
        vector we start at the given `point`, which is expected to be on the line.

        Args:
            point: Point to start from. Expected to be on the line.
            vector: Vector to travel along.
            min_distance: Minimum distance that the new point should have with respect to the line. For clarity,
                this is not the same as the distance traveled along the vector.
            collinearity_epsilon_deg: Collinearity epsilon in degrees, to make sure the vector and the line
                are not collinear. If their angle is less than this value, a ValueError is raised.

        Raises:
            ValueError: If the vector and the line are collinear, according to the given epsilon.

        Returns:
            A tuple with the distance traveled along the vector, and the new point at that distance.
        """
        vector = vector.normalized()

        # Make sure they are not parallel, to prevent div/0.
        ori = Orientation.of_vec_with_respect_to(
            vec=vector, wrt=self.direction, collinearity_epsilon_degrees=collinearity_epsilon_deg
        )
        if ori.is_collinear():
            raise ValueError("The vector and the line direction are parallel; cannot achieve MIN_D.")

        # Calculate sin(theta) using the cross product
        sin_theta_abs = abs(self.direction.cross(vector))

        # Calculate the distance to move along the vector, and return the new point at that distance.
        desired_d = min_distance / sin_theta_abs
        p_along_vec = point + vector * desired_d
        return desired_d, p_along_vec



from __future__ import annotations

import heapq
import itertools
import math
from enum import Enum, auto
from typing import Generic, Iterator, Protocol, TypeAlias, TypeVar, cast

from py_straight_skeleton.constants import DISTANCE_EPSILON, NONCOLLINEARITY_EPSILON_DEGS, TIME_EPSILON
from py_straight_skeleton.polygon_math import LAV, DegenerateLAVError, Edge, VertexInfo, is_point_inside_polygon_edges
from py_straight_skeleton.skeleton import Skeleton
from py_straight_skeleton.vector_math import Line2, Orientation, Ray2, Vector2


class AlgorithmTracerProtocol(Protocol):
    """Protocol to receive notifications from key algorithm steps and edge cases."""

    def on_reflex_vertex_found_after_edge_event(self, vertex: VertexInfo):
        """Called when we decide a vertex is reflex after an edge event, which should not happen and causes issues."""
        ...

    def on_failed_to_find_opposite_edge(self, event_info: SplitEventInfo):
        """Called when we fail to find the opposite edge for a split event, which sometimes is expected if the
        opposite edge has collapsed due to other events."""
        ...

    def on_step_begin(self, step: int, time: float, algorithm: StraightSkeletonAlgorithm):
        """Called at the beginning of each iteration of the algorithm's main loop."""
        ...

    def on_event_popped(self, step: int, time: float, event_info: EventInfo, algorithm: StraightSkeletonAlgorithm):
        """Called when an event is popped from the priority queue and is about to be processed."""
        ...

    def on_vertex_attempting_event_requeue(self, vertex: VertexInfo):
        """Called when a vertex that failed to process an event is attempting to requeue a new event."""
        ...

    def on_vertex_became_its_opposite_edge(self, event_info: SplitEventInfo):
        """Called when a vertex becomes its own opposite edge, which should not happen and can cause issues."""
        ...

    def on_algorithm_end(self, algorithm: StraightSkeletonAlgorithm):
        """Called when the algorithm finishes processing all events and the skeleton is complete."""
        ...

    def on_algorith_raised_exception(self, algorithm: StraightSkeletonAlgorithm, exception: Exception):
        """Called when the algorithm raises an exception, which should not happen and signals a bug."""
        ...

    def on_debug_hit(self, algorithm: StraightSkeletonAlgorithm):
        """General hook for debugging during edge cases."""
        ...


class NoopAlgorithmTracer(AlgorithmTracerProtocol):
    """Implementation of the AlgorithmTracerProtocol that does nothing on each notification."""

    def on_reflex_vertex_found_after_edge_event(self, vertex: VertexInfo):
        pass

    def on_failed_to_find_opposite_edge(self, event_info: SplitEventInfo):
        pass

    def on_step_begin(self, step: int, time: float, algorithm: StraightSkeletonAlgorithm):
        pass

    def on_event_popped(self, step: int, time: float, event_info: EventInfo, algorithm: StraightSkeletonAlgorithm):
        pass

    def on_vertex_attempting_event_requeue(self, vertex: VertexInfo):
        pass

    def on_vertex_became_its_opposite_edge(self, event_info: SplitEventInfo):
        pass

    def on_algorithm_end(self, algorithm: StraightSkeletonAlgorithm):
        pass

    def on_algorith_raised_exception(self, algorithm: StraightSkeletonAlgorithm, exception: Exception):
        pass

    def on_debug_hit(self, algorithm: StraightSkeletonAlgorithm):
        pass


# Global tracer that can be replaced by the user to receive key events for debugging and logging.
GLOBAL_ALGORITHM_TRACER: AlgorithmTracerProtocol = NoopAlgorithmTracer()


def set_global_algorithm_tracer(tracer: AlgorithmTracerProtocol):
    """Set a custom global algorithm tracer for algorithm notifications."""
    global GLOBAL_ALGORITHM_TRACER
    GLOBAL_ALGORITHM_TRACER = tracer


T = TypeVar("T")


class PriorityQueue(Generic[T]):
    """Priority queue implementation using a heap for efficient insertion and removal of items with priorities."""

    def __init__(self):
        """Initialize a new empty Priority Queue."""
        self._data = []

    def __len__(self) -> int:
        """Return the number of items in the priority queue."""
        return len(self._data)

    def __iter__(self) -> Iterator[tuple[float, T]]:
        """Iterate over the elements in priority order without modifying the queue, using a sorted copy."""
        return iter(sorted(self._data))

    def put(self, priority: float, item: T) -> None:
        """Insert the given item in the queue with the given priority.

        Args:
            priority: The priority of the item. Lower values are better priorities.
            item: The item to insert in the queue.
        """
        heapq.heappush(self._data, (priority, item))

    def get(self) -> tuple[float, T]:
        """Remove and return the item with the lowest(=best) priority from the queue. Raises IndexError if empty."""
        return heapq.heappop(self._data)

    def peek(self) -> tuple[float, T]:
        """Return the item with the lowest(=best) priority without removing it. Raises IndexError if empty."""
        if len(self._data) == 0:
            raise IndexError("Attempted to peek from empty priority queue.")
        return self._data[0]

    def empty(self) -> bool:
        """Return True if the priority queue is empty, False otherwise."""
        return len(self._data) == 0


class EventInfo:
    """Information about an 'event' in the context of edge wavefront events."""

    class EventType(Enum):
        EDGE = auto()  # An edge collapses.
        SPLIT = auto()  # A reflex vertex splits an edge.

    __id_counter = itertools.count()

    @classmethod
    def reset_count(cls):
        cls.__id_counter = itertools.count()

    def __init__(self, event_type: EventType, position: Vector2, distance: float):
        """Initialize the event with the given type, position and distance. A new unique id is automatically assigned
        to the event.

        Args:
            event_type: Type of event.
            position: Point where the event happens (generally a collision point between wavefronts).
            distance: Distance to the vertex that detected the event.
        """
        self._auto_id = next(self.__id_counter)

        self.event_type = event_type
        self.position = position
        self.distance = distance

    def __lt__(self, other: EventInfo):
        """Compare events based on their distance first. If the distance is the same (within epsilon), compare the auto
        id of the events, thus prioritizing the event that was detected first."""

        if not isinstance(other, EventInfo):
            return NotImplemented

        # TODO: In older versions of the algorithm, before event requeing on event failure, it was necessary to sort by
        # distance here if the priorities were tied. It seems that it now causes issues, and that event detection time
        # (equal to auto_id) is preferred in order to break ties. This however is brittle, and requires better
        # decision making when events happen at the same priority.
        # if abs(self.distance - other.distance) < TIME_EPSILON:
        #     return self._auto_id < other._auto_id
        # else:
        #     return self.distance < other.distance
        return self._auto_id < other._auto_id

    @property
    def is_edge_event(self):
        return self.event_type == self.EventType.EDGE

    @property
    def is_split_event(self):
        return self.event_type == self.EventType.SPLIT


class EdgeEventInfo(EventInfo):
    """Information about an edge event, where two bisectors intersect to collapse an edge."""

    def __init__(self, position: Vector2, distance: float, v_a: VertexInfo, v_b: VertexInfo):
        """Initialize the edge event with the given intersection position, the distance to the vertex that detected the
        event, and the two vertices whose bisectors cause the intersection, and therefore collapse of edge."""
        super().__init__(event_type=self.EventType.EDGE, position=position, distance=distance)
        self.v_a = v_a
        self.v_b = v_b

    def __repr__(self):
        return f"EdgeEventInfo({self._auto_id}, d={self.distance}, pos={self.position}, v_a={self.v_a}, v_b={self.v_b})"


class SplitEventInfo(EventInfo):
    """Information about a split event, where a reflex vertex splits an edge."""

    def __init__(self, position: Vector2, distance: float, v: VertexInfo, opposite_edge: Edge):
        """Initialize the split event with the given split position, the distance to the vertex causing the split, such
        vertex, and the original edge that will be split."""
        super().__init__(event_type=self.EventType.SPLIT, position=position, distance=distance)
        self.v = v
        self.opposite_edge = opposite_edge

    def __repr__(self):
        return f"SplitEventInfo({self._auto_id}, d={self.distance}, pos={self.position}, v={self.v}, edge={self.opposite_edge})"


# Type definitions.
EventQueue: TypeAlias = PriorityQueue[EventInfo]
SLAV: TypeAlias = set[LAV]
PointList: TypeAlias = list[tuple[float, float]]


class StraightSkeletonAlgorithm:
    """Encapsulates the straight skeleton computation algorithm itself."""

    def __init__(self, event_time_decimals: int | None = 6):
        """Initialize the algorithm with the given number of decimal places for event priorities.

        Args:
            event_time_decimals: The number of decimal places to consider to round event priorities and distances. If
                None, full precision is used. If set, the priorities and the distance to vertice will be rounded to the
                given number of decimals before they are compared. Note that if the number of decimals is less than the
                time comparison epsilon, rounding can cause issues, since rounding can pull close times apart more than
                that epsilon (eg: if decimals is 1, two close events 0.47 and 0.51 would now be 0.1 apart, which is much
                greater than the original 0.04 difference). This value should probably be >= decimals(TIME_EPSILON) + 1.
        """
        self.event_priority_decimals = event_time_decimals
        self.original_edges: list[Edge] = []
        self.skeleton: Skeleton = Skeleton(original_edges=[])
        self.slav = SLAV()

    def _round(self, value: float | None) -> float | None:
        """Round and return the given value to the number of decimals set for the algorithm."""
        if value is None:
            return None
        return round(value, self.event_priority_decimals) if self.event_priority_decimals is not None else value

    def _compute_closest_edge_event(self, vi: VertexInfo) -> tuple[float, EdgeEventInfo] | tuple[None, None]:
        """Computes the edge event which is closest to the given vertex, if any. Edge events happen when the bisector
        of the given vertex and the bisector of one of its neighbors intersect, collapsing the edge between them.

        Args:
            vi: The vertex for which to compute the nearest edge event.

        Returns:
            A tuple with the priority of the event and the event information, if an event is found. If no event is
            found, the tuple (None, None) is returned. The priority is the distance from the intersection point to the
            vanishing edge's holding line.
        """
        # Compute the intersection of the current bisector with the prev and next bisectors.
        cur_bisector = vi.bisector_ray()
        prev_bisector = vi.prev_vertex.bisector_ray()
        next_bisector = vi.next_vertex.bisector_ray()
        inter_prev = cur_bisector.intersection_with_ray(prev_bisector)
        inter_next = cur_bisector.intersection_with_ray(next_bisector)
        # Check if both intersections are None.
        if inter_prev is None and inter_next is None:
            return None, None
        # Take vars from the closer intersection.
        if inter_next is None:
            loc_i, va, vb = inter_prev, vi.prev_vertex, vi
        elif inter_prev is None:
            loc_i, va, vb = inter_next, vi, vi.next_vertex
        else:
            len_prev = (inter_prev - vi.position).length()
            len_next = (inter_next - vi.position).length()
            # If they are the same either should be fine.
            loc_i, va, vb = (
                (inter_prev, vi.prev_vertex, vi) if len_prev < len_next else (inter_next, vi, vi.next_vertex)
            )

        # Pick the edge (should be either next(a) or prev(b)).
        if va.next_orig_edge != vb.prev_orig_edge:
            raise RuntimeError("Unexpected condition: next_edge(a) != prev_edge(b)")
        edge_i = va.next_orig_edge

        # Compute queue priority and create the event.
        priority = edge_i.distance_point_to_edge_line(point=loc_i)
        priority = self._round(value=priority)
        distance = (loc_i - vi.position).length()
        distance = self._round(value=distance)
        event_info = EdgeEventInfo(position=loc_i, distance=distance, v_a=va, v_b=vb)
        return priority, event_info

    def _compute_split_events(self, vi: VertexInfo) -> None:
        """Compute the split events for a given vertex, and cache them in the vertex. Split events happen when a reflex
        vertex splits an opposing edge, potentially splitting the polygon into separate subpolygons.

        Args:
            vi: The vertex for which to compute the nearest split event.

        Raises:
            RuntimeError: In case of unexpected computations/conditions (might signal a bug).
        """
        if not vi.is_reflex:
            vi.pending_split_events = []
            return

        # Grab the segments for the vertex that we are going to operate on.
        try:
            prev_segment = (vi.position - vi.get_prev_vertex_safe().position).normalized()
            next_segment = (vi.get_next_vertex_safe().position - vi.position).normalized()
        except DegenerateLAVError:
            # LAV is a point, will not attempt to find split event.
            vi.pending_split_events = []
            return

        found_split_events = []

        # Note the order is important since events with same time will be ordered according to detection time.
        v_bisector_ray = vi.bisector_ray()
        for edge in self.original_edges:
            # Skip the edges for the vertex itself, they can't be the opposite edge.
            if edge.has_endpoint(vi):
                continue
            edge_line = Line2(point=edge.start_p, direction=edge.direction)

            # Check the intersection between the vertex's bisector and the line the edge lies on. If there is no
            # intersection, the edge is "behind" the vertex.
            inter_bisector_ei_line = edge_line.intersection_with_ray(ray=v_bisector_ray)
            if inter_bisector_ei_line is None:
                continue

            # Check the vertex's segments (as ending on it).
            for segment in [prev_segment, -next_segment]:
                # If the segment is collinear to the tested edge, skip it.
                if Orientation.of_vec_with_respect_to(vec=segment, wrt=edge.direction).is_collinear():
                    continue

                # Compute the "axis of the angle" between the segment and the edge line, choosing the direction of the
                # edge given the orientation of the axis with respect to the vertex's bisector.
                #   1. Compute the intersection point between the edge line and the segment line.
                #   2. Compute the axis of the angle by selecting the edge direction according to the orientation.
                #   3. Compute Bi as the intersection between the axis and the vertex's own bisector.
                segment_line = Line2(point=vi.position, direction=segment)

                # 1. Compute the intersection point between the edge line and the segment line.
                int_on_edge = edge_line.intersection_with_line(segment_line)
                if int_on_edge is None:
                    raise RuntimeError("Intersection expected since the lines are not parallel.")

                if vi.position == int_on_edge:
                    # The intersection is the vertex itself, so the vertex is part of the edge line.
                    # I think we can break instead of continue, since the other segment will not be valid either, but
                    # would need to verify that theory.
                    continue

                # 2. Compute the axis of the angle by selecting the edge orientation according to where the intersection
                # point lies with respect to the vertex's bisector.
                int_to_v_dir = (vi.position - int_on_edge).normalized()
                orientation = Orientation.of_vec_with_respect_to(
                    vec=int_to_v_dir,
                    wrt=v_bisector_ray.direction,
                    collinearity_epsilon_degrees=NONCOLLINEARITY_EPSILON_DEGS,
                    # collinearity_epsilon_degrees=TIGHT_COLLINEARITY_EPSILON_DEGS, See note below in `Issue: Collinear`
                )

                # Special consideration for collinear cases.
                if not orientation.is_collinear():
                    # If the incoming vector is to the right of the bisector, we invert the edge, since the intersection
                    # point is to the left of the vertex with respect to its bisector.
                    edge_dir = -edge.direction if orientation.is_right_exclusive() else edge.direction
                    int_bisector = (edge_dir + int_to_v_dir).normalized()

                    # 3. Compute Bi as the intersection between the axis and the vertex's own bisector.
                    int_b = v_bisector_ray.intersection_with_ray(Ray2(int_on_edge, int_bisector))
                    if int_b is None:
                        # There is no intersection, there will be no candidate point.
                        break
                else:
                    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                    # Issue: Collinear. There is a substantial problem here. The point we are computing is at an
                    # arbitrary location, since the event wants to happen almost on top of the edge itself. The
                    # correction that we do may, when events are processed, land very close to another skeleton node,
                    # while having a slightly different time, since they can come from different edges. During skeleton
                    # verification, this can result in nodes being at the same location, while having too different
                    # times, which is not correct (and currently fires an exception). To solve the issue we could:
                    # 0) Only provide one point, and crash due to the precision issue.
                    # a) Set `collinearity_epsilon_degrees` to NONCOLLINEARITY_EPSILON_DEGS above, so that we truly do
                    #   not fall here ever and trust precision to do its job.
                    # b) Flag in the event that the point is arbitrary, and that we are open to moving it.
                    # c) Create more than one event with different locations, flagging them as linked, and letting
                    #   one and only one contribute to the skeleton.
                    # ... others?
                    # Although I think (c) is the most correct, I need time to implement it and test it, so we will go
                    # with (a) for now. (b) seems wrong, since moving the point essentially causes different times,
                    # which could have impacted the skeleton construction up to that point.
                    # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                    # The code below is for (0), which will currently not fire due to (a) being set.
                    # If the the vertex's bisector and its segment are almost parallel, Bi will be almost the same
                    # as the intersection point on the edge. However, the split point should not be on the edge, since
                    # that would be a degenerate case, so we need to find the closest point to the edge where Bi could
                    # be.
                    # In order to find where Bi could be, we know that its distance to the line holding the edge will be
                    # its priority, and that needs to be > TIME_EPSILON.
                    # MIN_DIST_B_TO_EDGE should be enough that the point is not collinear with respect to the edge, that
                    # depends on where the point is with respect to start_p, and the collinearity epsilon used in the
                    # region check `is_point_in_edge_region` below.
                    region_collinearity_eps = NONCOLLINEARITY_EPSILON_DEGS
                    min_degrees = region_collinearity_eps + 1e-5
                    dist_to_start_p = (int_on_edge - edge.start_p).length()
                    min_dist_b_to_edge = max(
                        math.sin(math.radians(min_degrees)) * dist_to_start_p, 2 * DISTANCE_EPSILON
                    )
                    # Compute such point.
                    desired_dist_along_v, desired_point = edge_line.find_point_with_min_distance_along_vector(
                        point=int_on_edge, vector=int_to_v_dir, min_distance=min_dist_b_to_edge
                    )
                    # But at the same time, we can't be past the vertex itself.
                    max_dist_int_towards_v = (vi.position - int_on_edge).length() - TIME_EPSILON

                    if desired_dist_along_v > max_dist_int_towards_v:
                        # If we can't find a point along the vector far enough from the edge but not past the vertex,
                        # we can't handle the split event that should go between the edge and the vertex.
                        raise RuntimeError("Reflex vertex too close to opposite edge without being part of the edge.")

                    # If we get here, we found a point.
                    int_b = desired_point

                # Now check whether Bi is inside the region defined by the edge and its bisectors.
                in_region = edge.is_point_in_edge_region(
                    point=int_b, collinearity_epsilon_degrees=NONCOLLINEARITY_EPSILON_DEGS
                )
                if not in_region:
                    # The candidate point is not in the region, this edge is not the opposite edge.
                    break

                # The point is a candidate event, add it the list of found events.
                dist_to_v = (int_b - vi.position).length()
                dist_to_v = self._round(value=dist_to_v)

                # Record the events.
                e_priority = edge.distance_point_to_edge_line(point=int_b)
                e_priority = self._round(value=e_priority)
                e_event_info = SplitEventInfo(position=int_b, distance=dist_to_v, v=vi, opposite_edge=edge)
                found_split_events.append((e_priority, e_event_info))

                # In any case, no need to check the second segment if the first one was not parallel, since B has the
                # same coordinates for both segments.
                break

        # Sort the split events by the distance to the vertex.
        found_split_events.sort(key=lambda x: x[1].distance)

        # Cache all split events in the vertex.
        vi.pending_split_events = found_split_events

    def _attempt_requeue_vertex_events(self, vi: VertexInfo) -> None:
        """Attempts to queue a new vertex event for the given vertex, which must be not yet processed.

        This is a modification to the original algorithm, which would only consider the closest event. When events fail
        (because part the other vertex has been processed or their opposite edge has collapsed), the unprocessed vetex
        can be left dangling. Those vertices have now a second chance of computing their next event. In order to do so,
        we compute the newest edge event possible, and check it against the next split event cached (if any).

        Args:
            vi: Vertex for which to attempt to queue a new event, since their intended event failed.

        Raises:
            RuntimeError: If the vertex is already processed.
        """
        if vi.processed:
            raise RuntimeError("Processed vertices don't get second chances.")

        # Notify that we are doing this to the tracer.
        GLOBAL_ALGORITHM_TRACER.on_vertex_attempting_event_requeue(vertex=vi)

        # Simply call queue again, since its implementation supports re-queueing events.
        self._queue_vertex_events(vi=vi)

    def _queue_vertex_events(self, vi: VertexInfo) -> None:
        """Compute the closest edge and all split events for the given vertex, and queue them with the appropriate
        priority, storing the cache of split events for later use if needed. If the cache was already computed, it will
        not be recomputed on subsequent calls.

        Args:
            vi: Vertex for which to compute the closest events.
        """
        # If already processed, this should not even be attempted.
        if vi.processed:
            raise RuntimeError("Processed vertices should not queue more events.")

        # Note the order of detection is important since events with same time will be ordered according to detection
        # time, as we compare their auto ids.

        # Check whether we should compute the split events now or they are already cached.
        if not vi.pending_split_events_computed:
            self._compute_split_events(vi=vi)

        # Always check the first split event against the edge event.
        if vi.pending_split_events:
            split_prio, split_event = vi.pending_split_events[0]
            split_dist = split_event.distance
        else:
            split_prio, split_event, split_dist = None, None, math.inf
        pop_split_event = False

        # Compute closest edge event.
        edge_prio, edge_event = self._compute_closest_edge_event(vi=vi)
        edge_dist = edge_event.distance if edge_event is not None else math.inf

        # Queue the closest one only.
        if split_dist < edge_dist:
            # Queue split event.
            assert split_prio is not None and split_event is not None  # for mypy
            self.queue.put(priority=split_prio, item=split_event)
            pop_split_event = True
        elif edge_dist < split_dist:
            # Queue edge event.
            assert edge_prio is not None and edge_event is not None  # for mypy
            self.queue.put(priority=edge_prio, item=edge_event)
        elif edge_event is not None:  # the distances are the same.
            # We can put both in the queue. Processing events at the same time is not a problem: we will process them
            # based on event comparison operator, and vertices that are already processed will be skipped.
            assert split_prio is not None and split_event is not None  # for mypy
            assert edge_prio is not None and edge_event is not None  # for mypy
            self.queue.put(split_prio, split_event)
            self.queue.put(edge_prio, edge_event)
            pop_split_event = True

        if pop_split_event:
            vi.pending_split_events.pop(0)

    def _initialize_from(self, exterior: PointList, holes: list[PointList]) -> None:
        """Initialize the algorithm's variables to compute the skeleton of a polygon defined by the given exterior
        and holes.

        Args:
            exterior: List of points defining the exterior of the polygon in COUNTER-CLOCKWISE orientation.
            holes: List of holes, each defined by a list of points in CLOCKWISE orientation.
        """
        self.original_edges = []
        self.slav = SLAV()
        self.queue = EventQueue()

        # Generate one LAV with the exterior vertices.
        first_lav = LAV()
        for v in exterior:
            # Skip last vertex for a closed polygon.
            p = Vector2(v[0], v[1])
            if first_lav.head is None or p != first_lav.head.position:
                first_lav.append_vertex_at(p)
        self.slav.add(first_lav)

        # Add holes (each as one lav).
        for hole in holes:
            hole_lav = LAV()
            for v in hole:
                p = Vector2(v[0], v[1])
                if hole_lav.head is None or p != hole_lav.head.position:
                    hole_lav.append_vertex_at(p)
            self.slav.add(hole_lav)

        # Build original edges from both exterior and holes.
        for lav in self.slav:
            for vi in lav:
                # Note the edge links itself to the vertices.
                self.original_edges.append(Edge(start_vertex=vi, end_vertex=vi.next_vertex))

        # Now we can create the skeleton instance, because we have computed the original edges.
        self.skeleton = Skeleton(original_edges=self.original_edges)

        # Compute each vertex info now that we have the edges.
        for lav in self.slav:
            for vi in lav:
                # Verify we set up the edges properly.
                assert vi.prev_orig_edge.end_vertex == vi, f"Vertex {vi} is not the end of its previous edge."
                assert vi.next_orig_edge.start_vertex == vi, f"Vertex {vi} is not the start of its next edge."
                # Compute bisector direction and reflexivity from the edges.
                vi.compute_from_edges()

        # Compute the initial set of events (edge and split) for each vertex in each lav.
        for lav in self.slav:
            for vi in lav:
                self._queue_vertex_events(vi=vi)

    def _process_edge_event(self, event_info: EdgeEventInfo, current_time: float) -> None:
        """Process an edge event, potentially adding arcs to the skeleton, and queueing new events if needed.

        Args:
            event_info: Event information.
            current_time: Current time.

        Raises:
            RuntimeError: If unexpected conditions are met (might signal a bug).
        """
        # (edge_event 2.b) If Va and Vb are processed, continue to (2).
        if event_info.v_a.processed or event_info.v_b.processed:
            if not event_info.v_a.processed:
                # v_a was not processed, but it's edge event vanished, give it a second chance for events.
                self._attempt_requeue_vertex_events(vi=event_info.v_a)

            if not event_info.v_b.processed:
                # v_b was not processed, but it's edge event vanished, give it a second chance for events.
                self._attempt_requeue_vertex_events(vi=event_info.v_b)

            # No further processing
            return

        # (edge_event 2.c) If prev(prev(Va)) == Vb (peak of the roof).
        if event_info.v_a.prev_vertex.prev_vertex == event_info.v_b:
            # Add three skeleton arcs [VaI, VbI, VcI], with Vc == prev(Va) = next(Vb), and continue to (2).
            if event_info.v_a.prev_vertex != event_info.v_b.next_vertex:
                raise RuntimeError("prev(prev(Va)) == Vb, but prev(Va) != next(Vb)")
            # Note we use a dummy here so that the skeleton can optimize finding its corresponding skeleton vertex.
            dummy_vertex = VertexInfo(event_info.position, lav_ptr=LAV())
            v_c = event_info.v_a.prev_vertex
            self.skeleton.add_arc(vertex_1=event_info.v_a, vertex_2=dummy_vertex, at_time=current_time)
            self.skeleton.add_arc(vertex_1=event_info.v_b, vertex_2=dummy_vertex, at_time=current_time)
            self.skeleton.add_arc(vertex_1=v_c, vertex_2=dummy_vertex, at_time=current_time)
            # Mark the three vertices as processed (although I think this is the end of the lav events).
            event_info.v_a.processed = True
            event_info.v_b.processed = True
            v_c.processed = True
            dummy_vertex.processed = True  # Also the dummy. Its lav is never part of the slav anyway.
            # We should remove this lav from the slav so the nodes are no longer searched for.
            self.slav.remove(event_info.v_a.lav_ptr)
            return

        # Implementation note: we do 2.e.2 before 2.d because we use the vertex's id as skeleton vertex id, so we need
        # to create the vertex before adding the arcs.
        # (edge_event 2.e.2) Create a new vertex at I.
        event_lav = event_info.v_a.lav_ptr
        new_vertex = VertexInfo(event_info.position, lav_ptr=event_lav)

        # (edge_event 2.d) Output two skeleton arcs [VaI, VbI].
        self.skeleton.add_arc(vertex_1=event_info.v_a, vertex_2=new_vertex, at_time=current_time)
        self.skeleton.add_arc(vertex_1=event_info.v_b, vertex_2=new_vertex, at_time=current_time)

        # (edge_event 2.e) Modify the LAV.
        # (edge_event 2.e.1) Mark Va and Vb as processed.
        event_info.v_a.processed = True
        event_info.v_b.processed = True
        # # (edge_event 2.e.2) Create a new vertex at I.
        # Done already (see note above).
        # event_lav = event_info.v_a.lav_ptr
        # new_vertex = VertexInfo(event_info.position, lav_ptr=event_lav)
        # (edge_event 2.e.3) Insert the new vertex into LAV, instead of Va and Vb.
        new_vertex.prev_vertex = event_info.v_a.prev_vertex
        new_vertex.next_vertex = event_info.v_b.next_vertex
        new_vertex.prev_vertex.next_vertex = new_vertex
        new_vertex.next_vertex.prev_vertex = new_vertex
        if event_lav.head == event_info.v_a or event_lav.head == event_info.v_b:
            event_lav.head = new_vertex
        # (edge_event 2.e.4) Link the appropriate edges to the new vertex.
        new_vertex.prev_orig_edge = event_info.v_a.prev_orig_edge
        new_vertex.next_orig_edge = event_info.v_b.next_orig_edge

        # - - - - - - - - - - - - - - - -
        # (edge_event 2.f) Compute V's events.
        # - - - - - - - - - - - - - - - -
        # (edge_event 2.f.1) Compute the new vertex's bisector.
        new_vertex.compute_from_edges()

        # TODO: This can still happen, but I believe it is correct to have them. I need to create a custom test to
        # verify that it is in fact required for them to be identified as reflex, so they can create split events even
        # if they are not original vertices. The test would involve two reflex vertices that collide, creating a
        # new vertex that is also reflex. Need to make sure that that reflex vertex should split the edge in front
        # of it. If so, then forcing that vertex to being non-reflex will be wrong.
        if new_vertex.is_reflex:
            new_vertex.force_reflex(value=False)
            GLOBAL_ALGORITHM_TRACER.on_reflex_vertex_found_after_edge_event(vertex=new_vertex)

        # (edge_event 2.f.2) Compute the intersections of the new bisector with the neighbor bisectors.
        # (edge_event 2.f.3) Store the nearer intersection point into the priority queue (if exists).
        # Implementation note: we would allow here split events if the vertex is reflex.
        self._queue_vertex_events(vi=new_vertex)

    def _process_split_event(self, event_info: SplitEventInfo, current_time: float) -> None:
        """Process a split event, potentially adding an arc to the skeleton, and queueing new events if needed.

        Args:
            event_info: Event information.
            current_time: Current time.

        Raises:
            RuntimeError: In case of unexpected conditions (might signal a bug).
        """
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # (split_event 2.b) if the intersection point points to already processed vertices continue on step 2.
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        if event_info.v.processed:
            return

        # (split_event 2.c) ... (continue)
        # (split_event 2.d) output one arc Vertex-Intersection of the straight skeleton.
        # Implementation note: This is a modification wrt to original algorithm, do not do it now since the split
        # event may fail due to the opposite edge having collapsed. In that case, we do not want to add the arc.
        # self.skeleton.add_arc(..., time=current_time) # <- not now

        # (split_event 2.e) modify the set of lists of active vertices/nodes (SLAV)
        # (split_event 2.e.1) Mark the vertex/node V (pointed to by I) as processed.
        # Implementation note: Modification wrt to original algorithm, do not do it now since the split event may fail.
        # event.v.processed = True # <- now now
        # (split_event 2.e.2) Create two new nodes V1 and V2 with the same coordinates as the intersection point I.
        tmp_lav = event_info.v.lav_ptr
        # Implementation note: we create a dummy vertex to hold the position of the split event in the skeleton,
        # and we make both nodes point to that dummy vertex, so that arcs in the skeleton all reference the same
        # vertex, even though they come from different LAVs and point at either vertex_v1 or vertex_v2.
        split_dummy_vertex = VertexInfo(position=event_info.position, lav_ptr=LAV())
        vertex_v1 = VertexInfo(position=event_info.position, lav_ptr=tmp_lav, split_dummy_vertex=split_dummy_vertex)
        vertex_v2 = VertexInfo(position=event_info.position, lav_ptr=tmp_lav, split_dummy_vertex=split_dummy_vertex)

        # (split_event 2.e.3) search the opposite edge in SLAV (sequentially).
        y, x = None, None
        for lavi in self.slav:
            # Iterate the lav finding the two actual vertices that now define the opposite edge.
            for cand_vertex in lavi:
                # Check if this vertex's right edge is the one we are searching for. If so, the next vertex's left
                # edge should also be the one we are searching for.
                if cand_vertex.next_orig_edge != event_info.opposite_edge:
                    continue

                # Sanity check that the edge is properly linked between two vertices.
                if cand_vertex.next_vertex.prev_orig_edge != event_info.opposite_edge:
                    raise RuntimeError("Next vertex's prev_edge is not the split opposite edge?")

                # Check whether the two sub-points of the opposite edge the ones in the event's subpoly. Note a tight
                # collinearity check is required here, or we risk degenerate results.
                y_cand = cand_vertex
                x_cand = cand_vertex.next_vertex
                if y_cand.position == event_info.position or x_cand.position == event_info.position:
                    # We created a split event some time ago, and now we have become part of the opposite edge's
                    # actual vertices. We can't process this.
                    # Note: this is a degenerate situation. Although it happened in the past, I think it is fixed now.
                    GLOBAL_ALGORITHM_TRACER.on_vertex_became_its_opposite_edge(event_info=event_info)
                    continue

                # Test that the point is in the region defined by the candidate split points.
                test_edge = Edge(start_vertex=y_cand, end_vertex=x_cand, is_original_edge=False)
                in_region = test_edge.is_point_in_edge_region(point=event_info.position)
                if not in_region:
                    continue

                if y == event_info.v or x == event_info.v:
                    # If the vertex that wanted to split the edge is one of the edge's vertices, we should not use it.
                    # Moreover, this is probably a bad event due to inclusive epsilon checks.
                    raise RuntimeError(f"Vertex {event_info.v} is one of the vertices of the opposite edge.")
                    # continue

                # Found
                y, x = y_cand, x_cand
                break

            # Do not check more lavs, we already found the vertices.
            if y is not None and x is not None:
                break

        # It is possible that we do not find the proper new vertices for the opposite edge, specifically when the edge
        # has collpased due to other events. If this happens, we should not flag the vertex as processed, so it can
        # have other events attempted.
        if y is None or x is None:
            GLOBAL_ALGORITHM_TRACER.on_failed_to_find_opposite_edge(event_info=event_info)
            # The opposite edge has collapsed, so we can't split it. But perhaps there are other that we can still
            # split. Give this vertex a second chance. This can often happen with reflex vertices that interrupt
            # the split events of other reflex vertices.
            self._attempt_requeue_vertex_events(vi=event_info.v)
            return

        # If we reach here, we can now process the event. Do the things we postponed earlier.
        self.skeleton.add_arc(vertex_1=event_info.v, vertex_2=split_dummy_vertex, at_time=current_time)
        event_info.v.processed = True

        # (split_event 2.e.4) insert both new nodes into the SLAV (break one LAV into two parts).
        # Implementation note: The paper does not describe what to do when the opposite edge is not part of the
        # lav that the event's vertex is part of. Empirically, this happens often in polygons with holes, and that
        # the expected result is to merge the two lavs into a single one. I also tested modifying the lavs as separate
        # ones and it did not yield good results, but would need to prove it.
        event_lav = event_info.v.lav_ptr
        is_multi_lav = event_lav != y.lav_ptr
        self.slav.remove(event_lav)

        # Insert nodes into the lav(s).
        if is_multi_lav:
            # Remove the lav where the points are, since we will merge it with the new lav.
            self.slav.remove(y.lav_ptr)
            # From two lavs, we merge them into a single one by connecting them with V1 and V2.
            #   [... M, Event, N ...]
            #   [... Y, X, ...      ]
            #                         -> [V1, X, ... Y, V2, N, ... M]
            # V1:
            vertex_v1.prev_vertex = event_info.v.prev_vertex
            vertex_v1.next_vertex = x
            # V2:
            vertex_v2.prev_vertex = y
            vertex_v2.next_vertex = event_info.v.next_vertex
            # Important: first set the pointers in both v1 and v2. Then set the pointers in the neighbors.
            vertex_v1.prev_vertex.next_vertex = vertex_v1
            vertex_v1.next_vertex.prev_vertex = vertex_v1
            vertex_v2.prev_vertex.next_vertex = vertex_v2
            vertex_v2.next_vertex.prev_vertex = vertex_v2
        else:
            # From one lav, we create two.
            # [X ... M, Event, N, ... Y]
            #                            -> [V1, X ... M]
            #                            -> [V2, N ... Y]
            # V1:
            vertex_v1.prev_vertex = event_info.v.prev_vertex
            vertex_v1.next_vertex = x
            # V2:
            vertex_v2.prev_vertex = y
            vertex_v2.next_vertex = event_info.v.next_vertex
            # Important: first set the pointers in both v1 and v2. Then set the pointers in the neighbors.
            vertex_v1.prev_vertex.next_vertex = vertex_v1
            vertex_v1.next_vertex.prev_vertex = vertex_v1
            vertex_v2.prev_vertex.next_vertex = vertex_v2
            vertex_v2.next_vertex.prev_vertex = vertex_v2

        # Create new lavs with updated lav pointers.
        new_lavs = []
        lav1 = LAV(head=vertex_v1)
        lav1_len = 0
        for v in lav1:
            v.lav_ptr = lav1
            lav1_len += 1
        new_lavs.append((lav1, lav1_len))

        creates_two_lavs = not is_multi_lav  # multi_lav input => single_lav output
        if creates_two_lavs:
            lav2 = LAV(head=vertex_v2)
            lav2_len = 0
            for v in lav2:
                v.lav_ptr = lav2
                lav2_len += 1
            new_lavs.append((lav2, lav2_len))

        # Add the new lavs, unless they collapse to a single line, in which case we can process immediately as a peak.
        for lavn, lavn_len in new_lavs:
            if lavn_len < 3:
                # If the new lav has less than 3 vertices, it is a line, and we should add it to the skeleton.
                lv1 = lavn.head.next_vertex
                lv2 = lavn.head  # This has to be either vertex_v1 or vertex_v2, so the given time is for that one.
                self.skeleton.add_arc(vertex_1=lv1, vertex_2=lv2, at_time=current_time)
                # Mark the two vertices as processed, since there may still be events in the queue.
                lv1.processed = True
                lv2.processed = True
            else:
                self.slav.add(lavn)

        # (split_event 2.e.5) link the new nodes V1 and V2 with the appropriate edges.
        vertex_v1.prev_orig_edge = vertex_v1.prev_vertex.next_orig_edge
        vertex_v1.next_orig_edge = vertex_v1.next_vertex.prev_orig_edge
        vertex_v2.prev_orig_edge = vertex_v2.prev_vertex.next_orig_edge
        vertex_v2.next_orig_edge = vertex_v2.next_vertex.prev_orig_edge

        # (split_event 2.f) Compute the events for the new nodes V1 and V2.
        # (split_event 2.f.1) First compute new angle bisectors between the line segments linked to them in step 2e.
        if not vertex_v1.processed:
            vertex_v1.compute_from_edges()
        if not vertex_v2.processed:
            vertex_v2.compute_from_edges()

        # (split_event 2.f.2 and split_event 2.f.3) Compute the events for the new nodes.
        for vertex in [vertex_v1, vertex_v2]:
            if not vertex.processed:
                self._queue_vertex_events(vi=vertex)

    def _compute_skeleton_impl(self, exterior: PointList, holes: list[PointList]) -> Skeleton:
        """Computes the straight skeleton of the polygon defined by the given exterior outline and interior holes.

        Args:
            exterior: List of vertices that define the exterior of the polygon. Must be provided in COUNTER-CLOCKWISE,
                with +X-right and +Y-up.
            holes: List of holes, each as a list of vertices that define the hole. Must be provided in CLOCKWISE order,
                with +X-right and +Y-up (opposite of the exterior vertices).

        Returns:
            The straight skeleton of the polygon defined by the exterior and holes.
        """
        # Reset all auto_id counters so that debugging ids are deterministic.
        VertexInfo.reset_count()
        Edge.reset_count()
        LAV.reset_count()
        EventInfo.reset_count()

        # Initial setup
        self._initialize_from(exterior=exterior, holes=holes)

        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # 2. While the priority queue is not empty do:
        # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        current_step = 0
        while not self.queue.empty():
            # Start the step.
            current_time = self.queue.peek()[0]
            current_step += 1
            GLOBAL_ALGORITHM_TRACER.on_step_begin(step=current_step, time=current_time, algorithm=self)

            # (2.a) Pop the lowest event from queue and process it.
            _, event_info = self.queue.get()
            GLOBAL_ALGORITHM_TRACER.on_event_popped(
                step=current_step, time=current_time, event_info=event_info, algorithm=self
            )

            # If the event is outside the polygon, ignore it. We should probably not queue it in the first place, but
            # leaving it here so we can raise an exception it if it happens.
            if not is_point_inside_polygon_edges(self.original_edges, point2d=event_info.position):
                # If the vertices have been processed, we can ignore the event.
                if event_info.is_edge_event:
                    edge_event_info = cast(EdgeEventInfo, event_info)
                    if edge_event_info.v_a.processed and edge_event_info.v_b.processed:
                        continue
                else:
                    split_event_info = cast(SplitEventInfo, event_info)
                    if split_event_info.v.processed:
                        continue
                # If the vertex is still unprocessed, this would generate a degenerate skeleton, so raise an error.
                # TODO This does happen in real cases. Originally I added it to detect unfinished skeletons, when
                # the last events were detected outside the polygon due to bad LAVs. However I have seen valid
                # cases in which a split event is generated outside the polygon due to the opposite edge collapsing
                # under earlier edge events. We also now detect the former cases during skeleton iteration.
                # raise RuntimeError("Event is still pending, but it is outside the polygon.")
                continue

            # Process the event.
            if event_info.is_edge_event:
                self._process_edge_event(event_info=cast(EdgeEventInfo, event_info), current_time=current_time)
            else:
                self._process_split_event(event_info=cast(SplitEventInfo, event_info), current_time=current_time)

            # We can stop processing events if there are no LAVs, since all vertices are now processed.
            if not self.slav:
                break

        GLOBAL_ALGORITHM_TRACER.on_algorithm_end(algorithm=self)

        return self.skeleton

    def compute_skeleton(self, exterior: PointList, holes: list[PointList]) -> Skeleton:
        """Computes the straight skeleton of the polygon defined by the given exterior outline and interior holes.

        Args:
            exterior: List of vertices that define the exterior of the polygon. Must be provided in COUNTER-CLOCKWISE,
                with +X-right and +Y-up.
            holes: List of holes, each as a list of vertices that define the hole. Must be provided in CLOCKWISE order,
                with +X-right and +Y-up (opposite of the exterior vertices).

        Returns:
            The straight skeleton of the polygon defined by the exterior and holes.
        """
        try:
            return self._compute_skeleton_impl(exterior=exterior, holes=holes)
        except Exception as e:
            GLOBAL_ALGORITHM_TRACER.on_algorith_raised_exception(algorithm=self, exception=e)
            raise e



from __future__ import annotations

import math
from collections import defaultdict
from typing import Iterator, TypeAlias

from py_straight_skeleton.constants import NONCOLLINEARITY_EPSILON_DEGS, TIME_EPSILON
from py_straight_skeleton.polygon_math import Edge, VertexInfo
from py_straight_skeleton.vector_math import Orientation, Vector2


class SkeletonNode:
    """Internal representation of vertices in the skeleton (aka nodes)."""

    def __init__(self, sk_id: int, position: Vector2, time: float):
        """Initialize a skeleton node with a given id, and its position and a time."""
        self._skn_id = sk_id
        self.position = position
        self.time = time

    def __repr__(self):
        return f"SKN({self._skn_id}, {self.position.x:.03f}, {self.position.y:.03f}, {self.time:.5f})"


# Undirected link between two nodes.
SkeletonArc: TypeAlias = tuple[SkeletonNode, SkeletonNode]

# List of indices. The indices are the node ids and also expected to be indexing into the list of all nodes in the
# skeleton.
SkeletonFace: TypeAlias = list[int]


class Skeleton:
    """Straight skeleton of a polygon. The skeleton is a graph of nodes and the arcs that connect them. Additionally,
    this class provides the time per node (=the distance from the original polygon edges), and computation of faces,
    so that the polygon can be triangulated and lifted.
    """

    def __init__(self, original_edges: list[Edge]):
        """Initialize the skeleton with the original edges of the polygon."""
        self._sk_nodes_all: list[SkeletonNode] = []
        self._sk_nodes_dummy_only: list[SkeletonNode] = []
        self._vertex_id_to_sk_node_id: dict[int, int] = {}  # This cache may be low hit, except at roof peaks.

        # _arcs_by_node: Bidirectional arcs by node id.
        self._arcs_by_node: dict[int, list[int]] = defaultdict(list)
        # _iteration_arc_list: Iterates arcs in one direction (creation direction), for example for plotting.
        self._iteration_arc_list: list[SkeletonArc] = []

        self._faces: list[SkeletonFace] | None = None

        # For every original edge, we create an initial skeleton node, and then we link them.
        self._sk_original_edges = []
        for edge in original_edges:
            self._add_skeleton_node(for_vertex=edge.start_vertex, at_time=0.0)
        for edge in original_edges:
            e_v_1 = self._find_skeleton_node(edge.start_vertex)
            e_v_2 = self._find_skeleton_node(edge.end_vertex)
            self._sk_original_edges.append((e_v_1._skn_id, e_v_2._skn_id))

    def _add_skeleton_node(self, for_vertex: VertexInfo, at_time: float) -> SkeletonNode:
        """Add an internal node from the given external vertex, at the given time.

        Args:
            vertex: External vertex to add to the skeleton.
            at_time: Time of the vertex, which is the distance from the original polygon edges.

        Returns:
            The newly created skeleton node.
        """
        next_id = len(self._sk_nodes_all)
        new_node = SkeletonNode(sk_id=next_id, position=for_vertex.position, time=at_time)
        self._sk_nodes_all.append(new_node)  # Possible optimization, do not build the list until finished.
        self._vertex_id_to_sk_node_id[for_vertex._auto_id] = new_node._skn_id
        return new_node

    def _find_skeleton_node(self, for_vertex: VertexInfo) -> SkeletonNode:
        """Find the skeleton node corresponding to the given external vertex, and raise an error if not found.

        Args:
            vertex: External vertex to find its corresponding skeleton node.

        Raises:
            RuntimeError: If the node is not found in the skeleton.

        Returns:
            The skeleton node corresponding to the given external vertex.
        """
        sk_vertex_id = self._vertex_id_to_sk_node_id.get(for_vertex._auto_id)
        if sk_vertex_id is not None:
            return self._sk_nodes_all[sk_vertex_id]
        raise RuntimeError(f"Expected vertex not found in skeleton: {for_vertex} vs \n{self._sk_nodes_all}")

    def _find_or_add_skeleton_node(self, for_vertex: VertexInfo, at_time: float) -> SkeletonNode:
        """Find the skeleton node corresponding to the given external vertex, or create a new one if not found.

        If the node is found, the time previously stored must match the given time, otherwise an error is raised.
        If the given time is 0, this method should not be called, and `_find_skeleton_node()` should be used instead,
        since the original vertices are expected to be found after construction, as long as original edges are properly
        specified.

        Args:
            vertex: External vertex to find or add to the skeleton.
            at_time: Time of the vertex, which is the distance from the original polygon edges.

        Raises:
            ValueError: If the time is zero, since you should call `_find_skeleton_node()` instead.
            RuntimeError: If the node is found but the given time does not match the stored time.

        Returns:
            The skeleton node corresponding to the given external vertex.
        """
        # If the time is zero, we should have expected the node to exist.
        if abs(at_time) <= TIME_EPSILON:
            raise ValueError("Zero time should have called `_find_skeleton_node()` instead.")

        # If a skeleton node was already present, we can avoid searching for it.
        sk_node_id = self._vertex_id_to_sk_node_id.get(for_vertex._auto_id)
        if sk_node_id is not None:
            return self._sk_nodes_all[sk_node_id]

        # Search in dummy nodes one that matches the position and the given time.
        for sk_dummy_node in self._sk_nodes_dummy_only:
            if sk_dummy_node.position == for_vertex.position:
                # We found a dummy at that location, check the times.
                if abs(sk_dummy_node.time - at_time) <= TIME_EPSILON:
                    # Times match, we can return this one (and cache the mapping).
                    self._vertex_id_to_sk_node_id[for_vertex._auto_id] = sk_dummy_node._skn_id
                    return sk_dummy_node
                else:
                    # Times do not match, this is an error, since the positions overlap and thus require the same time.
                    raise RuntimeError(
                        f"Nodes at the same position but different time: {for_vertex} vs {sk_dummy_node}\n"
                        + f"  new time:{at_time} vs old time:{sk_dummy_node.time} (diff={abs(sk_dummy_node.time - at_time)})\n"
                        + f"  new position     : {for_vertex.position.x}, \t{for_vertex.position.y}\n"
                        + f"  vs node position : {sk_dummy_node.position.x}, \t{sk_dummy_node.position.y}\n"
                        + f"  diff             : {abs(for_vertex.position.x-sk_dummy_node.position.x)}, \t{abs(for_vertex.position.y-sk_dummy_node.position.y)}"
                    )

        # We did not find one, we have to create a new dummy node (and cache in the dummy list).
        new_dummy_node = self._add_skeleton_node(for_vertex=for_vertex, at_time=at_time)
        self._sk_nodes_dummy_only.append(new_dummy_node)
        return new_dummy_node

    def add_arc(self, vertex_1: VertexInfo, vertex_2: VertexInfo, at_time: float) -> None:
        """Add an arc between the given external vertices, at the given time. Note that the time is a reference to the
        time of the second vertex, and that the first vertex must already exist in the skeleton, either as an original
        vertex or as a previously added vertex.

        Args:
            vertex_1: External vertex to connect.
            vertex_2: External vertex to connect.
            at_time: Time at which the second vertex happens.
        """
        # If any of the vertices are from a split event, link to their dummy vertex instead.
        if vertex_1.split_dummy_vertex is not None:
            vertex_1 = vertex_1.split_dummy_vertex
        if vertex_2.split_dummy_vertex is not None:
            vertex_2 = vertex_2.split_dummy_vertex

        # Find the skeleton nodes for the given vertices.
        sk_node_1 = self._find_skeleton_node(for_vertex=vertex_1)
        sk_node_2 = self._find_or_add_skeleton_node(for_vertex=vertex_2, at_time=at_time)
        skn_id_1, skn_id_2 = sk_node_1._skn_id, sk_node_2._skn_id

        # If they are the same, we should not add an arc, since it's a self referencing arc.
        if skn_id_1 == skn_id_2:
            # We could log this, although it is expected when two edges collapse at the same meeting point.
            return

        # Now add the connection between the nodes.
        self._arcs_by_node[skn_id_1].append(skn_id_2)
        self._arcs_by_node[skn_id_2].append(skn_id_1)
        self._iteration_arc_list.append((sk_node_1, sk_node_2))

    def _compute_faces_if_needed(self) -> None:
        """Compute the faces that the polygon is divided into by the skeleton, and store them in the instance.

        Raises:
            RuntimeError: If the faces could not be computed, for example if the skeleton is not a valid skeleton graph.
        """
        if self._faces is not None:
            return

        # Sanity check that each node has its index as it's node id.
        for idx, sk_v in enumerate(self._sk_nodes_all):
            if sk_v._skn_id != idx:
                raise RuntimeError(f"Node id mismatch: {sk_v._skn_id} vs {idx}")

        faces = []
        # Each original edge causes a face. We can iterate the skeleton to find loops.
        # o1 -> o2 -> sk1 ... skn -> o1
        # Options/Heuristics:
        #   1) DFS, in which loops ending at a visited node or an original other than o1 are discarded.
        #   2) Follow the smallest internal angle at each node. <-- We'll do this.
        for skn_id_1, skn_id_2 in self._sk_original_edges:
            face: SkeletonFace = []

            # Grab the two nodes of the original edge.
            node1, cur_node = self._sk_nodes_all[skn_id_1], self._sk_nodes_all[skn_id_2]
            face.append(node1._skn_id)

            # Iterate the next possible nodes and pick the smallest interior angle.
            prev_node = node1
            while cur_node != node1:
                face.append(cur_node._skn_id)
                # Find the next node with the smallest interior angle (they are left angles, so minimize cosine=dot).
                cur_dir = (cur_node.position - prev_node.position).normalized()
                next_node = None
                min_cos = math.inf
                for next_id in self._arcs_by_node[cur_node._skn_id]:
                    # We can't traverse the same arc back.
                    if next_id == prev_node._skn_id:
                        continue
                    if next_id == cur_node._skn_id:
                        raise RuntimeError(f"Found arc to self: {cur_node}")
                    # Check the direction of this arc.
                    cand_node = self._sk_nodes_all[next_id]
                    cand_dir = (cand_node.position - cur_node.position).normalized()
                    cand_ori = Orientation.of_vec_with_respect_to(
                        vec=cand_dir, wrt=cur_dir, collinearity_epsilon_degrees=NONCOLLINEARITY_EPSILON_DEGS
                    )
                    if cand_ori.is_opposite():
                        # Opposite, we should not have this.
                        raise RuntimeError(f"Found opposite arc to non-prev node: {cur_node} -> {cand_node}")
                    # If this arc has the smallest angle so far, keep it.
                    cos_angle = cur_dir.dot(cand_dir)
                    if cand_ori.is_right_exclusive():
                        cos_angle = 2.0 - cos_angle  # For concave, re-scale the cosine [-1, 1] -> [3, 1]
                    if cos_angle < min_cos:
                        min_cos = cos_angle
                        next_node = cand_node
                # We should have found a new node, and it should not be the current one.
                if next_node is None or next_node == cur_node:
                    raise RuntimeError(f"Failed to find a proper next node building faces: {next_node}")
                if next_node == node1:
                    # We have closed the loop, we are done.
                    face.append(node1._skn_id)
                    break
                # The new node should not be an original one, since that means that we did not close the loop.
                if next_node.time <= TIME_EPSILON:
                    raise RuntimeError(f"Ran to an original node {next_node} not part of the initiating edge.")
                prev_node = cur_node
                cur_node = next_node

            # Add the face to the list of faces.
            faces.append(face)

        # We are done, store the faces.
        self._faces = faces

    @property
    def nodes(self) -> list[SkeletonNode]:
        """Get all the nodes in the skeleton, including both polygon-original vertices and additional nodes."""
        return self._sk_nodes_all

    def get_faces(self) -> list[SkeletonFace]:
        """Get the faces that the polygon is divided into by the skeleton, as a list of internal node ids. Note that
        faces will be computed if they have not been cached yet."""
        # Compute faces if not done yet.
        self._compute_faces_if_needed()
        assert self._faces is not None  # For mypy
        return self._faces

    def arc_iterator(self) -> Iterator[tuple[SkeletonNode, SkeletonNode]]:
        """Iterate the arcs in the skeleton, as pairs of nodes that are connected by an arc. Note that this iterator
        iterates the arcs directionally only once, despite the arcs not having a real direction."""
        return iter(self._iteration_arc_list)



"""
Utilities and tracers to plot the state of the straight skeleton algorithm to Matplotlib figures.
"""

import logging
import os
from typing import Any, Sequence

from matplotlib import pyplot as plt
from matplotlib.figure import Figure
from matplotlib.patches import FancyArrowPatch
from matplotlib.typing import ColorType
from py_straight_skeleton.algorithm import EventInfo, EventQueue, NoopAlgorithmTracer, StraightSkeletonAlgorithm
from py_straight_skeleton.polygon_math import (
    LAV,
    Edge,
    NoopPolygonMathTracer,
    VertexInfo,
    VertexNotComputedError,
    is_point_inside_polygon_edges,
)
from py_straight_skeleton.skeleton import Skeleton

logger = logging.getLogger(__name__)


class Utils:
    """Utilities for plotting elements of the straight skeleton algorithm."""

    @staticmethod
    def plot_vertex_info(
        vi: VertexInfo,
        ax: plt.Axes,
        *,
        bisector_color: ColorType | None = None,
        bisector_length: float = 1,
        arrow_color: ColorType | None = None,
        hide_original_vertex_arrows: bool = True,
    ) -> None:
        """Plot the given vertex information.

        Args:
            vi: Vertex information.
            ax: Matplotlib axes.
            bisector_color: Color of the bisector. Bisector will not be shown if None.
            bisector_length: Length of the bisector.
            arrow_color: Color of the arrows that point towards the vertex's connected edges. Arrows will not be shown
                if None.
            hide_original_vertex_arrows: Whether to hide the arrows that point towards the vertex's connected edges if
                the vertex is defined by both its edges, as is the case of the original vertices.
        """
        # Vertex bisector.
        if bisector_color is not None:
            try:
                bis_direction = vi.bisector_direction
                ax.plot(
                    [vi.position.x, vi.position.x + bis_direction.x * bisector_length],
                    [vi.position.y, vi.position.y + bis_direction.y * bisector_length],
                    color=bisector_color,
                )
            except VertexNotComputedError:
                pass

        # Vertex arrows to its connected edges.
        if arrow_color is not None:
            skip_arrow = (
                hide_original_vertex_arrows
                and vi.prev_orig_edge.has_endpoint(vi)
                and vi.next_orig_edge.has_endpoint(vi)
            )
            if not skip_arrow:
                ax.add_patch(
                    FancyArrowPatch(
                        vi.position.coords,
                        vi.prev_orig_edge.center_p.coords,
                        connectionstyle="arc3,rad=.5",
                        color=arrow_color,
                        arrowstyle="->",
                        mutation_scale=20,
                    )
                )
                ax.add_patch(
                    FancyArrowPatch(
                        vi.position.coords,
                        vi.next_orig_edge.center_p.coords,
                        connectionstyle="arc3,rad=-.5",
                        color=arrow_color,
                        arrowstyle="->",
                        mutation_scale=20,
                    )
                )

    @staticmethod
    def plot_lav(
        lav: LAV, ax: plt.Axes, *, show_bisectors=True, bisector_length: float = 1, show_edge_arrows: bool = True
    ):
        """Plots a LAV.

        Args:
            lav: Lav to plot.
            ax: Matplotlib axes.
            show_bisectors: Whether to show the bisectors of the vertices in the LAV.
            bisector_length: Length of the vertex bisectors. Ignored if `show_bisectors` is False.
            show_edge_arrows: Whether to show the arrows that point towards the vertex's connected edges.
        """
        xs, ys, colors = [], [], []
        for i, vi in enumerate(lav):
            xs.append(vi.position.x)
            ys.append(vi.position.y)
            try:
                colors.append("red" if vi.processed else "magenta" if vi.is_reflex else "goldenrod")
            except VertexNotComputedError:
                colors.append("black")

            if not vi.processed:
                bisector_color = colors[-1] if show_bisectors else None
                arrow_color = (0.8, 0.8, 0.8) if show_edge_arrows else None
                Utils.plot_vertex_info(
                    vi=vi,
                    ax=ax,
                    bisector_color=bisector_color,
                    bisector_length=bisector_length,
                    arrow_color=arrow_color,
                )

        # Plot the vertices.
        ax.scatter(xs, ys, c=colors, s=6)
        # Plot the segments.
        ax.plot(xs + [xs[0]], ys + [ys[0]], color=(0.1, 0.1, 0.1), linestyle="-", lw=1)

        # Annotate the vertices.
        for i, vi in enumerate(lav):
            # Draw odd to the left, even to the right.
            if i % 2 == 0:
                ax.text(vi.position.x, vi.position.y, f"{i}", fontsize=7, ha="left", va="bottom")
            else:
                ax.text(vi.position.x, vi.position.y, f"{i}", fontsize=7, ha="right", va="top")

    @staticmethod
    def plot_edges(edges: list[Edge], ax: plt.Axes, show_edge_ids: bool = False, **kwargs) -> None:
        """Plots the given edges to the given Matplotlib axes, passing through the kwargs."""
        for edge in edges:
            ax.plot([edge.start_p.x, edge.end_p.x], [edge.start_p.y, edge.end_p.y], **kwargs)
            if show_edge_ids:
                ax.text(edge.center_p.x, edge.center_p.y, f"e{edge._auto_id}.", fontsize=7)

    @staticmethod
    def plot_queue(queue: EventQueue, ax: plt.Axes, original_edges: list[Edge] | None):
        """Plots each event in the given queue.

        Args:
            queue: Queue to plot.
            ax: Matplotlib axes.
            original_edges: Original edges of the polygon. If provided, events outside the polygon will not be plotted.
        """
        # Iterate the queue without removing elements
        for idx, (_, event_info) in enumerate(queue):
            cc = event_info.position
            if original_edges and not is_point_inside_polygon_edges(edges=original_edges, point2d=cc):
                logger.debug(f"Not plotting event {idx} @ {cc} as it is outside the polygon.")
                continue
            if idx == 0:
                color = "red"
                marker = "o" if event_info.is_edge_event else "p"
                dests = (
                    [event_info.v_a.position, event_info.v_b.position]
                    if event_info.is_edge_event
                    else [event_info.v.position, event_info.opposite_edge.center_p]
                )
                for dest in dests:
                    debug_arrow = FancyArrowPatch(
                        cc.coords,
                        dest.coords,
                        color=color,
                        arrowstyle="->",
                        mutation_scale=20,
                        connectionstyle="arc3,rad=-.5",
                    )
                    ax.add_patch(debug_arrow)
                s = 64
            elif idx == 1:
                color = "orange"
                marker = "." if event_info.is_edge_event else "*"
                s = 16
            else:
                color = "blue" if event_info.is_edge_event else "magenta"
                marker = "d" if event_info.is_edge_event else "x"
                s = 12
            ax.scatter(cc.x, cc.y, color=color, marker=marker, s=s)

    @staticmethod
    def plot_skeleton(
        skeleton: Skeleton,
        ax: plt.Axes,
        *,
        arc_color: ColorType = "darkgreen",
        circle_color: ColorType | None = None,
        show_vertex_info: bool = False,
        vertex_info_fontsize: int = 6,
        face_colors: Sequence[ColorType] | Sequence[str] | ColorType | None = None,
        **kwargs,
    ):
        """Plots the given skeleton to the given Matplotlib axes

        Args:
            skeleton: Skeleton to plot.
            ax: Matplotlib axes to plot to.
            color: Color of the skeleton arcs.
            circle_color: Color of the circles to render at each vertex. Circles no rendered if set to None.
            show_vertex_info: If set to true, the information (id and time) of each vertex will be displayed.
            vertex_info_fontsize: Font size of the vertex information. Ignored if `show_vertex_info` is False.
            face_colors: Color or colors to render the faces of the skeleton. If a single color is provided, all faces
                will be rendered with that color. If a list of colors is provided, the faces will be rendered with the
                colors in the list, repeating if necessary. If None, faces will not be rendered. Note that this will
                cause face computation if the skeleton does not have faces computed yet.
        """
        # Render the faces first (so that arcs and info overwrite them).
        if face_colors is not None:
            face_color_l = [face_colors] if not hasattr(face_colors, "__getitem__") else face_colors  # type: ignore
            for face_idx, face in enumerate(skeleton.get_faces()):
                face_color = face_color_l[face_idx % len(face_color_l)]
                face_coords = [(skeleton.nodes[v_idx].position.x, skeleton.nodes[v_idx].position.y) for v_idx in face]
                ax.fill(*zip(*face_coords), color=face_color, alpha=0.5)
        # Render the arcs next.
        for skv1, skv2 in skeleton.arc_iterator():
            ax.plot([skv1.position.x, skv2.position.x], [skv1.position.y, skv2.position.y], color=arc_color, **kwargs)
            # Render a circle at p2 with the time as its radius
            if circle_color is not None:
                ax.add_patch(
                    plt.Circle((skv2.position.x, skv2.position.y), radius=skv2.time, color=circle_color, fill=False)
                )
        # Show vertex information at each vertex last (if desired).
        if show_vertex_info:
            for sk_v in skeleton.nodes:
                x, y = sk_v.position.x, sk_v.position.y
                ax.text(x, y, f"{sk_v._skn_id} | t={sk_v.time:.2f}", fontsize=vertex_info_fontsize)


class PlotTracer(NoopAlgorithmTracer, NoopPolygonMathTracer):
    """Tracer that plots the state of the algorithm and other events to Matplotlib figures."""

    def __init__(
        self,
        *,
        fig_cs: float = 5.0,
        step_fig_nrows: int = 3,
        step_fig_ncols: int = 3,
        step_max_figs: int = 3,
        output_dir: str | None = None,
        log_level: int | str = logging.DEBUG,
    ):
        """Initialize the tracer with custom Matplotlib figure settings. The settings for figures are applied to
        step plotting, but some may not affect other plots, such as final plots on algorithm end.

        Args:
            fig_cs: Width=Height of the cells in inches. See figsize in plt.subplots.
            step_fig_nrows: Number of rows per figure before creating another for step plots.
            step_fig_ncols: Number of columns per figure before creating another for step plots.
            step_max_figs: Maximum number of figures to create for step plots. If 0 (or negative), no figures will be
                created.
            output_dir: If set, figures will also be saved to files automatically.
            log_level: Logging level to use for the tracer. It is applied to the module's logger.
        """
        super().__init__()

        # General.
        self.fig_cs = fig_cs

        # Step auto-figures.
        self.step_fig_nrows = step_fig_nrows
        self.step_fig_ncols = step_fig_ncols
        self.step_max_figs = step_max_figs
        self.step_cur_fig: tuple[str, Figure] | None = None
        self.step_next_ax_index = 0
        self.step_cur_flat_axes: Any | None = None

        # Output folder.
        self.output_dir = output_dir
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)

        # Logging.
        logger.setLevel(log_level)

    def _create_new_step_figure_lazy(self, fig_name: str) -> bool:
        """Create a new Matplotlib figure if the next axis index is out of bounds, and we have not reached the limit.

        Args:
            fig_name: Name of the figure to create (if one is created), to identify it when we save to disk.

        Returns:
            True if we can use the next axis, False if we have reached the limit of figures and we did not create
            a new one.
        """
        # If our next ax is still within the current figure, we don't need to do anything.
        if self.step_cur_flat_axes is not None and self.step_next_ax_index < len(self.step_cur_flat_axes):
            return True

        # Otherwise, save the current one. It will be closed if it is saved.
        self._save_current_figure_if_needed(close_it=True)

        # If we have reached the limit, we can't create a new one.
        if self.step_max_figs <= 0:
            return False

        # Create a new one.
        fig, axes = plt.subplots(
            ncols=self.step_fig_ncols,
            nrows=self.step_fig_nrows,
            figsize=(self.fig_cs * self.step_fig_ncols, self.fig_cs * self.step_fig_nrows),
        )
        fig.tight_layout()

        # Configure it.
        self.step_cur_fig = (fig_name, fig)
        self.step_next_ax_index = 0
        self.step_cur_flat_axes = axes.flatten() if self.step_fig_nrows > 1 else [axes]
        for ax in self.step_cur_flat_axes:
            ax.set_aspect("equal")

        # And discount from figures left.
        self.step_max_figs -= 1
        return True

    def _get_next_ax(self, fig_name: str) -> plt.Axes | None:
        """Return a new axis to use, increasing the index for the next one and creating a new figure if necessary.

        Args:
            fig_name: Name of the created figure (if one is created), to identify it when we save to disk.

        Returns:
            The axis to plot to, or None if we have reached the maximum number of figures allowed.
        """
        can_draw = self._create_new_step_figure_lazy(fig_name=fig_name)
        if not can_draw:
            return None

        assert self.step_cur_flat_axes is not None  # for mypy
        ax = self.step_cur_flat_axes[self.step_next_ax_index]
        self.step_next_ax_index += 1
        return ax

    def _save_current_figure_if_needed(self, *, close_it: bool) -> None:
        """Save to disk the current figure if we have a folder set up, and a figure to save.

        Args:
            close_it: Whether to close the figure after saving it. In general this should be True, but sometimes
                we may want to save a figure due to a debug event, while keeping it open for more plots.
        """
        if self.step_cur_fig is not None and self.output_dir is not None:
            fig_name, fig = self.step_cur_fig
            target_dir = os.path.join(self.output_dir, f"skel_plot_{fig_name}.png")
            fig.savefig(target_dir)
            if close_it:
                plt.close(fig)
            logger.debug(f"Saved plot to {target_dir}")

        # Regardless of whether we saved the image, remove the pointers as if we had closed it. We do not close it
        # so that it stays open for interactive notebooks, but we pretend we did so that we create new ones later on.
        if close_it:
            self.step_cur_fig = None
            self.step_cur_flat_axes = None
            self.step_next_ax_index = 0

    def _plot_common_to_ax(
        self,
        ax: plt.Axes,
        algorithm: StraightSkeletonAlgorithm,
        *,
        bisector_length: float | None = None,
        show_edge_ids: bool = False,
    ) -> None:
        """Plot most common elements of the algorithm to the given axis.

        Args:
            ax: Axis to plot to.
            algorithm: Instance with the current state of the algorithm.
            bisector_length: Length of the vertex bisectors. If None, a default length is computed, based on a factor
                with respect to the axis size.
            show_edge_ids: Whether to show the edge ids.
        """

        def compute_bisector_length(bisector_factor: float = 0.5) -> float:
            """Helper to compute bisector length as a factor of the axis size."""
            x_range = ax.get_xlim()
            y_range = ax.get_ylim()
            x_width = abs(x_range[1] - x_range[0])
            y_height = abs(y_range[1] - y_range[0])
            return min(x_width, y_height) * bisector_factor

        bisector_length = compute_bisector_length() if bisector_length is None else bisector_length

        # Original edges
        Utils.plot_edges(
            edges=algorithm.original_edges, ax=ax, color="gray", linestyle="--", show_edge_ids=show_edge_ids
        )
        # LAVs.
        for lav in algorithm.slav:
            Utils.plot_lav(lav=lav, ax=ax, bisector_length=bisector_length, show_edge_arrows=False)
        # Skeleton
        Utils.plot_skeleton(skeleton=algorithm.skeleton, ax=ax, arc_color="lightsteelblue", circle_color=None)
        # Queue
        Utils.plot_queue(queue=algorithm.queue, ax=ax, original_edges=algorithm.original_edges)

    def on_step_begin(self, step: int, time: float, algorithm: StraightSkeletonAlgorithm) -> None:
        """Logs the current queue and plots the state of the algorithm if we have axes/plots left to plot to."""

        # Log current queue.
        logger.debug("= " * 80)
        logger.debug("Queue:")
        for idx, (priority, event_info) in enumerate(algorithm.queue):
            logger.debug(f"{idx} ({event_info._auto_id}): {priority} {event_info.event_type:15s} {event_info.position}")
        logger.debug("= " * 80)

        # Grab the next axis to plot to.
        fig_name = f"step_{step}"
        cur_step_ax = self._get_next_ax(fig_name=fig_name)
        if cur_step_ax is None:
            return

        # Set ax's title.
        queue_count = len(algorithm.queue)
        lav_count = len(algorithm.slav)
        cur_step_ax.set_title(f"Step={step}, t={time:.6f}, {queue_count} events, {lav_count} LAVs", fontsize=8)

        # Plot everything.
        self._plot_common_to_ax(ax=cur_step_ax, algorithm=algorithm, show_edge_ids=step == 1)

    def on_event_popped(
        self, step: int, time: float, event_info: EventInfo, algorithm: StraightSkeletonAlgorithm
    ) -> None:
        """Logs the event information."""

        # Print event information.
        logger.debug(
            f"s={step}, t={time: .6f}. Processing event at {event_info.position} (type: {event_info.event_type})"
        )

    def on_algorithm_end(self, algorithm: StraightSkeletonAlgorithm) -> None:
        """Closes the current figure, and plots the final result of the algorithm, polygonizing the skeleton."""

        # Add to the limit of figs, since we will need at least two more axes.
        self.step_max_figs += 2

        # Skeleton
        final_ax_skel = self._get_next_ax(fig_name="end_skeleton")
        assert final_ax_skel is not None  # for mypy
        final_ax_skel.set_aspect("equal")
        final_ax_skel.set_title("Final Skeleton", fontsize=8)
        Utils.plot_edges(edges=algorithm.original_edges, ax=final_ax_skel, color="gray", linestyle="--")
        Utils.plot_skeleton(skeleton=algorithm.skeleton, ax=final_ax_skel, circle_color="whitesmoke")

        # Polygonized skeleton.
        final_ax_faces = self._get_next_ax(fig_name="end_faces")
        assert final_ax_faces is not None  # for mypy
        final_ax_faces.set_aspect("equal")
        final_ax_faces.set_title("Final Faces", fontsize=8)

        # Plot end result.
        Utils.plot_edges(edges=algorithm.original_edges, ax=final_ax_faces, color="gray", linestyle="--")
        face_colors = ["red", "green", "blue", "yellow", "purple", "orange", "cyan", "magenta"]
        try:
            # Polygonization can be tricky, log any issuies, but do not crash while plotting.
            Utils.plot_skeleton(
                skeleton=algorithm.skeleton, ax=final_ax_faces, arc_color="black", face_colors=face_colors, lw=0.5
            )
        except Exception as e:
            logger.error(f"Failed to plot skeleton faces: {e}")

        # Immediately save/close the figure if needed.
        self._save_current_figure_if_needed(close_it=True)

    def on_lav_infinite_loop_detected(self, lav: LAV) -> None:
        """Save the last plot we were able to get, since we are crashing."""

        logger.error("Infinite loop detected in LAV computation.")

        # Save the last image since we are crashing.
        self._save_current_figure_if_needed(close_it=True)

    def on_debug_hit(self, algorithm: StraightSkeletonAlgorithm) -> None:
        """General function to plot the current state of the algorithm when a debug hit is triggered."""

        # Create a debug figure and plot all to it.
        debug_fig, debug_ax = plt.subplots(figsize=(10, 10))
        debug_ax.set_title("Debug hit")
        debug_ax.set_aspect("equal")

        # Plot common state of the algorithm.
        self._plot_common_to_ax(ax=debug_ax, algorithm=algorithm)

        # - PLACEHOLDER BEGIN
        # Here we can add custom information during debugging.
        # - PLACEHOLDER END

        # Save it to the output folder.
        if self.output_dir:
            target_dir = os.path.join(self.output_dir, "on_debug_hit.png")
            debug_fig.savefig(target_dir)
            plt.close(debug_fig)
            logger.info(f"Saved debug hit plot to {target_dir}")

        # Save the current figure if needed, but do not close it, in case the algorithm continues.
        self._save_current_figure_if_needed(close_it=False)

    def on_algorith_raised_exception(self, algorithm: StraightSkeletonAlgorithm, exception: Exception) -> None:
        """Save the last plot we were able to get, since we are crashing."""

        # Save the last image since we are crashing.
        self._save_current_figure_if_needed(close_it=True)

        # Also attempt to draw a last image.
        self.on_debug_hit(algorithm=algorithm)
